import numpy as np
from allensdk.ephys.feature_extractor import EphysFeatureExtractor
import allensdk.ephys.ephys_extractor as efex
import allensdk.ephys.ephys_features as ft

ALIGN_CUT_WINDOW = np.array([ 0.002, 0.015 ])

def find_spikes_list_old(voltage_list, dt):
    out_idx = []
    out_v = []
    
    for v in voltage_list:
        idx, v = find_spikes_old(v, dt)
        out_idx.append(idx)
        out_v.append(v)

    return out_idx, out_v

def find_spikes_list(voltage_list, dt):
    v_set = [ v * 1e3 for v in voltage_list ]
    t_set = [ np.arange(0, len(v)) * dt for v in voltage_list ]
    i_set = [ np.zeros(len(v)) for v in voltage_list ]
    
    ext = efex.EphysSweepSetFeatureExtractor(t_set, v_set, i_set, filter=None)
    ext.process_spikes()
    sweep_spikes = [ s.spikes() for s in ext.sweeps() ]
    
    out_idx = [ np.array([ int(s['threshold_index']) for s in spikes ]) for spikes in sweep_spikes ]
    out_v = [ np.array([ s['threshold_v'] for s in spikes ]) for spikes in sweep_spikes ]
    
    return out_idx, out_v

SHORT_SQUARE_MAX_THRESH_FRAC = 0.1

def find_spikes_ssq_list(voltage_list, dt, dv_cutoff, thresh_frac):
    v_set = [ v * 1e3 for v in voltage_list ]
    t_set = [ np.arange(0, len(v)) * dt for v in voltage_list ]
    i_set = [ np.zeros(len(v)) for v in voltage_list ]

    thresh_frac = max(SHORT_SQUARE_MAX_THRESH_FRAC, thresh_frac)
    
    ext = efex.EphysSweepSetFeatureExtractor(t_set, v_set, i_set, 
                                             dv_cutoff=dv_cutoff, 
                                             thresh_frac=thresh_frac,
                                             filter=None)
    ext.process_spikes()
    sweep_spikes = [ e.spikes() for e in ext.sweeps() ]
    
    out_idx = [ np.array([ int(s['threshold_index']) for s in spikes ]) for spikes in sweep_spikes ]
    out_v = [ np.array([ s['threshold_v'] for s in spikes ]) for spikes in sweep_spikes ]
    
    return out_idx, out_v

def find_spikes_old(v, dt):
    v = v * 1e3 # convert V => mV
    t = np.arange(0, len(v)) * dt    
    i = np.zeros(t.shape)
    
    
    
    fx = efex.EphysSweepFeatureExtractor(t=t, v=v, i=i)
    fx.process_spikes()
    feature_data = fx.spikes()
    
    #fx = EphysFeatureExtractor()
    #fx.process_instance("", v, i ,t, 0, t[-1], "")
    #feature_data = fx.feature_list[0].mean

    ids = np.array([ s["threshold_idx"] for s in feature_data ])
    vs = np.array([ s["threshold_v"] for s in feature_data ]) 

    vs /= 1e3 # mV => V

    return ids, vs


def align_and_cut_spikes(voltage_list, current_list, dt, spike_window = None):
    ''' This function aligns the spikes to some criteria and returns a current and voltage trace of 
    of the spike over a time window.  Also returns zero crossing,and threshold 
    in reference to the aligned spikes.
    '''
    if spike_window is None:
        spike_window = ALIGN_CUT_WINDOW

    spike_shapes = []
    current_shapes = []
    index_before_spike = int(spike_window[0] / dt)
    index_after_spike = int(spike_window[1] / dt)
    aligned_spike_ind = np.array([])
    spike_sweeps = []
    spikes_per_trace = np.array([])
    
    spike_ind_list, _ = find_spikes_list(voltage_list, dt)

    for jj, voltage_and_current_and_spike in enumerate(zip(voltage_list, current_list, spike_ind_list)):
        voltage, current, whole_trace_spike_ind = voltage_and_current_and_spike
        
        spikes_per_trace = np.append(spikes_per_trace, len(whole_trace_spike_ind))
        
        alignment_ind = whole_trace_spike_ind                        
        aligned_spike_ind = np.append(aligned_spike_ind, np.ones(len(whole_trace_spike_ind)) * index_before_spike)

        # print 'alignment_ind', alignment_ind
        spike_delimiters = [(ind - index_before_spike, ind + index_after_spike) for ind in alignment_ind]
        for d in spike_delimiters: 
            # this 'if' statement makes sure we don't cause a ValueError
            if min(d) > 0 and max(d) < len(voltage) - 1:
                spike_trace = voltage[d[0]:d[1]]
                current_trace = current[d[0]:d[1]]
                spike_shapes.append(spike_trace)
                current_shapes.append(current_trace)                  
                spike_sweeps.append(jj)

            
    # note: that depending on how things were aligned, all of one of the values will be the same.
    print "spikes_per_trace", spikes_per_trace
    temp = np.append(0, np.cumsum(spikes_per_trace))  
    print 'temp', temp 
    wave_index_of_first_spikes = [int(ii) for ii in list(temp[range(0, len(temp) - 1)])]         
    print "in cut spikes: wave_index_of_first_spikes ", wave_index_of_first_spikes

    return spike_shapes, current_shapes, aligned_spike_ind, wave_index_of_first_spikes, spike_sweeps



#going to need to take preprocessed dictionaries and model configuration and create 
#a preprocessed model configuration and preprocessed_config file
#
#something will have to tell it what parameters to take out of the preprocessed dict
import os
import sys
import logging
import time
import numpy as np
import argparse
import allensdk.core.json_utilities as ju
import find_sweeps as fs

class ModelConfigurationException( Exception ): pass

DEFAULT_NEURON_PARAMETERS = {
    "type": "GLIF", 
    "dt": 5e-05, 
    "El": 0,
    "asc_tau_array": [ 1, 1 ],
    "asc_amp_array": [ 0, 0 ],
    "init_AScurrents": [ 0.0, 0.0 ],
    "init_threshold": 0.02, 
    "init_voltage": 0.0,
    "extrapolation_method_name": "endpoints",
    "dt_multiplier": 1
    }

DEFAULT_OPTIMIZER_PARAMETERS = { 
    "xtol": 1e-05, 
    "ftol": 1e-05, 
    "sigma_outer": 0.3, 
    "sigma_inner": 0.01, 
    "inner_iterations": 3,  
    "outer_iterations": 3, 
    "internal_iterations": 10000000,
    "iteration_info": [],
    "param_fit_names": [],
    "cut": 0,
    "bessel": { 'N': 4, 'freq': 10000 }
    }
    

def specify_parameter_groups(dictionary, dict_specifer, neuron_type):
    '''Specifies which values from the preprocessor will be used in the model configuration. 
    This is helpful if the preprocessor calculates many different values. 
        
    Parameters
    ----------
    dictionary: dict
        dictionary from preprocessor
    dict_specifier: string
        The following are available model levels 
        'LIF' (GLIF1)
        'LIF_R' (GLIF2)
        'LIF_ASC' (GLIF3)
        'LIF_R_ASC'  (GLIF_4) 
        'LIF_R_ASC_AT'  (GLIF_5)
    neuron_type: string
        'simple_neuron' is the only available option however here would be a good place for 
        the user to implement their own configurations.
    
    Returns
    -------
    output_dict: dict
        dictionary containing model configuration
    '''

    output_dict={'El_reference':dictionary['El']['El_noise']['measured']['mean'],
             'El':0.,
             'dt':dictionary['dt_used_for_preprocessor_calculations'], 
             'spike_cut_length':dictionary['spike_cutting']['NOdeltaV']['cut_length'],
             'spike_cutting_intercept':dictionary['spike_cutting']['NOdeltaV']['intercept'],
             'spike_cutting_slope':dictionary['spike_cutting']['NOdeltaV']['slope'],
             'asc_amp_array':dictionary['asc']['amp'],
             'asc_tau_array':(1./np.array(dictionary['asc']['k'])).tolist(),
             'th_inf': dictionary['th_inf']['via_Vmeasure']['from_zero'],
             'deltaV': None,
             'threshold_adaptation': {'a_spike_component_of_threshold': dictionary['threshold_adaptation']['a_spike_component_of_threshold'],
                                      'b_spike_component_of_threshold':dictionary['threshold_adaptation']['b_spike_component_of_threshold'],
                                      'a_voltage_component_of_threshold':dictionary['threshold_adaptation']['a_voltage_comp_of_thr_from_fitab'],
                                      'b_voltage_component_of_threshold': dictionary['threshold_adaptation']['b_voltage_comp_of_thr_from_fitab']},
             'MLIN': dictionary['MLIN'],
             'spike_inds': {
                'noise1': [ ],
                'noise2': [ ]
              }
             } 
    
    # specify specific values different for different levels.  Although there is only one neuron 
    # type here, this would be a good place to add other user defined neuron types
    if neuron_type=='simple_neuron':
        output_dict['C']=dictionary['capacitance']['C_test_list']['mean']
        if dict_specifer in ['LIF', 'LIF_R']:
            output_dict['R_input']=dictionary['resistance']['R_test_list']['mean']        
        elif dict_specifer in ['LIF_ASC', 'LIF_R_ASC', 'LIF_R_ASC_AT']:
            output_dict['R_input']=dictionary['resistance']['R_fit_ASC_and_R']['mean'] 
 
    for k,v in dictionary['sweep_properties']['noise1'].iteritems():
        output_dict['spike_inds']['noise1'].append( v['spike_ind'] )
        output_dict['spike_inds']['noise2'].append( v['spike_ind'] )
                
    return output_dict


def validate_method_requirements(method_config_name, has_mss):
    '''Confirm that the neuron has the specific sweeps required for the specified configuration
        
    Parameters
    ----------
    method_config_name: string
        Specifies the model level. Options are: 
            'LIF' (GLIF1)
            'LIF_R' (GLIF2)
            'LIF_ASC' (GLIF3)
            'LIF_R_ASC'  (GLIF_4) 
            'LIF_R_ASC_AT'  (GLIF_5)
    has_mss: boolean
        Specifies if the neuron has a multi short square sweep (for fitting spike component of threshold).
    '''
    if not has_mss:
        valid_configs = ['LIF', 'LIF_ASC']  
    else:
        valid_configs = ['LIF', 'LIF_ASC','LIF_R', 'LIF_R_ASC', 'LIF_R_ASC_AT']

    if method_config_name not in valid_configs:
        raise ModelConfigurationException("Model type %s cannot be configured due to missing data (mss: %s)" % ( method_config_name, str(has_mss)))

def update_neuron_method(method_type, arg_method_name, neuron_config):
    #TODO: documentation
    neuron_config[method_type] = { 'name': arg_method_name, 'params': None }


def configure_model(method_config, preprocessor_values):
    '''Configures the model from the specified method configuration and preprocessor values.
    
    Parameters
    ----------
    method_config:  dictionary
        contains values needed to configure the methods for the specified level within the dictionary
    preprocessor_values: dictionary
        dictionary from preprocessor
    '''
    
    preprocessor_values = specify_parameter_groups(preprocessor_values, method_config['name'], 'simple_neuron')

    neuron_config = {}
    neuron_config.update(DEFAULT_NEURON_PARAMETERS)
    optimizer_config = {}
    optimizer_config.update(DEFAULT_OPTIMIZER_PARAMETERS)

    #a) select values want to use out of the preprocessor_values via specifying parameter_gropus
    #b) look what levels are available via the levels available in the preprocessor_values.
    
    # Skip trace if subthreshold noise has a spike in it.
    noise1_ind =  [ n1i for n1i in preprocessor_values['spike_inds']['noise1'] if n1i is not None ]
    noise1_ind = np.concatenate(noise1_ind)
    if np.any(noise1_ind * preprocessor_values['dt'] < 8.0):
        raise ModelConfigurationException("Subthreshold region of noise1 stimulus contains spikes.")

    # check if there is a short square triple
    if preprocessor_values['threshold_adaptation']['b_spike_component_of_threshold'] and preprocessor_values['threshold_adaptation']['a_spike_component_of_threshold']:
        has_mss=True
    else:
        has_mss=False
        
    # make sure that the requested method config meets minimum requirements
    validate_method_requirements(method_config['name'], has_mss)

    update_neuron_method('AScurrent_dynamics_method', method_config['AScurrent_dynamics_method'], neuron_config)
    update_neuron_method('voltage_dynamics_method', method_config['voltage_dynamics_method'], neuron_config)
    update_neuron_method('threshold_dynamics_method', method_config['threshold_dynamics_method'], neuron_config)
    update_neuron_method('AScurrent_reset_method', method_config['AScurrent_reset_method'], neuron_config)
    update_neuron_method('voltage_reset_method', method_config['voltage_reset_method'], neuron_config)
    update_neuron_method('threshold_reset_method', method_config['threshold_reset_method'], neuron_config)

    neuron_config['El_reference'] = preprocessor_values['El_reference']
    neuron_config['C'] = preprocessor_values['C']           
    neuron_config['El'] = preprocessor_values['El']  
    neuron_config['spike_cut_length'] = preprocessor_values['spike_cut_length']
    neuron_config['asc_amp_array'] = preprocessor_values['asc_amp_array']
    neuron_config['asc_tau_array'] = preprocessor_values['asc_tau_array']
    neuron_config['R_input'] = preprocessor_values['R_input']
    neuron_config['th_inf'] = preprocessor_values['th_inf']

    optimizer_config['error_function'] = method_config['error_function']
    optimizer_config['param_fit_names'] = method_config['param_fit_names']

    #b) choose the sets want from the preprocessor_values 
    configure_method_parameters(neuron_config,
                                optimizer_config,
                                preprocessor_values['spike_cutting_slope'],
                                preprocessor_values['spike_cutting_intercept'],
                                preprocessor_values['threshold_adaptation']['a_spike_component_of_threshold'],
                                preprocessor_values['threshold_adaptation']['b_spike_component_of_threshold'],
                                preprocessor_values['threshold_adaptation']['a_voltage_component_of_threshold'],
                                preprocessor_values['threshold_adaptation']['b_voltage_component_of_threshold'],
                                preprocessor_values['MLIN']['var_of_section'],
                                preprocessor_values['MLIN']['sv_for_expsymm'],
                                preprocessor_values['MLIN']['tau_from_AC'])

    return {
        'neuron': neuron_config,
        'optimizer': optimizer_config
        }

def configure_method_parameters(neuron_config,
                                optimizer_config,
                                v_reset_slope,
                                v_reset_intercept,
                                a_spike_component_of_threshold,
                                b_spike_component_of_threshold,
                                a_voltage_component_of_threshold,
                                b_voltage_component_of_threshold,
                                var_of_section,
                                sv_for_expsymm,
                                tau_from_AC):
    '''Configures the methods used to run the models
    
    Parameters
    ----------
    neuron_config: dict
        contains neuron parameters
    optimizer_config: dict
        contains parameters for optimizaton
    v_reset_slope: float
        slope of the line in voltage reset
    v_reset_intercept: float
        intercept of the line in voltage reset
    a_spike_component_of_threshold: float or None
        amplitude of spike component of the threshold
    b_spike_component_of_threshold: float or None
        time course of spike component of the threshold
    a_voltage_component_of_threshold: float or None
        a parameter in voltage component of threshold
    b_voltage_component_of_threshold: float or None
        b parameter in voltage component of threshold
    var_of_section: float
        variance in noise of highest amplitude subthreshold long square pulse
    sv_for_expsymm: float
        parameter in MLIN optimization
    tau_from_AC: float
        time course of exponential fit to the autocorrelation
    '''
    
    # configure voltage reset rules
    method_config = neuron_config['voltage_reset_method']
    if method_config.get('params', None) is None:
        if method_config['name'] == 'zero':
            method_config['params'] = {}
        elif method_config['name'] == 'v_before':                        
            method_config['params'] = {
                'a': v_reset_slope,
                'b': v_reset_intercept
                }
    
        elif method_config['name'] == 'i_v_before':
            method_config['params'] = {
                'a': 1,
                'b': 2,
                'c': 3
                }
            raise ModelConfigurationException('i_v_before of voltage reset method is not yet implemented')
        elif method_config['name'] == 'fixed':
            raise ModelConfigurationException('cannot use fixed voltage reset method in preprocessor')
        else:
            method_config['params'] = {}
    
    # configure threshold reset rules  
    method_config = neuron_config['threshold_reset_method']        
    if method_config.get('params', None) is None:
        coeff_th_inf = neuron_config.get('coeffs', {}).get('th_inf',1.0)
        adjusted_th_inf = neuron_config['th_inf'] * coeff_th_inf
            
        if method_config['name'] == 'max_v_th':
            raise ModelConfigurationException('max_v_th threshold reset rule is not currently in use')
            
        elif method_config['name'] == 'th_before':
            raise ModelConfigurationException('th_before is not currently in use')
            
        elif method_config['name'] == 'inf':
            method_config['params'] = {}
            neuron_config['init_threshold'] = adjusted_th_inf
                
        elif method_config['name'] == 'three_components':
            method_config['params'] = { 'a_spike': a_spike_component_of_threshold,
                                       'b_spike': b_spike_component_of_threshold }
            neuron_config['init_threshold'] = adjusted_th_inf
                
        elif method_config['name'] == 'fixed':
            raise ModelConfigurationException("cannot use fixed threshold reset method in preprocessor")
        else:  
            raise ModelConfigurationException("unknown threshold reset method: ", method_config['name'])
    
    # configure voltage dynamics rules
    
    method_config = neuron_config['voltage_dynamics_method']
    if method_config.get('params', None) is None:
        if method_config['name'] == 'quadratic_i_of_v':
            raise ModelConfigurationException('quadraticIofV of voltage_dynamics_method preprocessing is not yet implemented')
        elif method_config['name'] == 'linear_forward_euler':
            method_config['params'] = {}
        elif method_config['name'] == 'linear_exact':
            method_config['params'] = {}

        else:
            raise ModelConfigurationException("unknown voltage dynamics method: ", method_config['name'])
    
    # configure threshold dynamics rules    
    method_config = neuron_config['threshold_dynamics_method']
    
    if method_config.get('params', None) is None:
        if method_config['name'] == 'three_components_forward':
            method_config['params'] = {
                    'a_spike': a_spike_component_of_threshold,
                    'b_spike': b_spike_component_of_threshold,
                    'a_voltage': a_voltage_component_of_threshold,
                    'b_voltage': b_voltage_component_of_threshold
                    }
            
        elif method_config['name'] == 'three_components_exact':
            method_config['params'] = {
                    'a_spike': a_spike_component_of_threshold,
                    'b_spike': b_spike_component_of_threshold,
                    'a_voltage': a_voltage_component_of_threshold,
                    'b_voltage': b_voltage_component_of_threshold
                    }

        elif method_config['name'] == 'spike_component':
            method_config['params'] = {
                    'a_spike': a_spike_component_of_threshold,
                    'b_spike': b_spike_component_of_threshold,
                    'a_voltage': 0,
                    'b_voltage': 0
                    }
    
        elif method_config['name'] == 'inf':
            method_config['params'] = {}
    
        else:
            raise ModelConfigurationException("unknown threshold dynamics method: ", method_config['name'])
    
    # configure ascurrent dynamics rules
    method_config = neuron_config['AScurrent_dynamics_method']
    if method_config.get('params', None) is None:
        # TODO: rename 'vector' to something more specific
        if method_config['name'] == 'vector': 
            method_config['params'] = {
                'vector': [1, 2, 3]
                }
            raise ModelConfigurationException('vector of AScurrent_dynamics_method is not yet implemented')  
        elif method_config['name'] == 'none':
            method_config['params'] = {}
        elif method_config['name'] == 'exp':
            method_config['params'] = {}
        else:
            raise ModelConfigurationException("unknown AScurrent dynamics method: ", method_config['name'])
    
    # configure ascurrent reset rule 
    # this is down here because it depends on numbers computed for the AScurrent_dynamics_method
    method_config = neuron_config['AScurrent_reset_method']
    if method_config.get('params', None) is None:
        if method_config['name'] == 'sum':
            method_config['params'] = {
                'r': np.ones(len(neuron_config['asc_tau_array']))
                }
        elif method_config['name'] == 'none':
            method_config['params'] = {}
        else:
            raise ModelConfigurationException("unknown AScurrent reset method: ", method_config['name'])
    
    # configure parameters for MLIN optimization
    if optimizer_config['error_function']=='MLIN':
        optimizer_config['error_function_data'] = {
            'subthreshold_long_square_voltage_variance': var_of_section,
            'sv_for_expsymm': sv_for_expsymm,
            'tau_from_AC': tau_from_AC
            }

    # validation

    # make sure that the initial ascurrents have the correct size
    if len(neuron_config['init_AScurrents']) != len(neuron_config['asc_tau_array']):
        raise ModelConfigurationException("init_AScurrents have incorrect length.")
    
    spike_cut_length = neuron_config.get('spike_cut_length', None)
    
    if spike_cut_length is None:
        raise ModelConfigurationException("Spike cut length must be set, but it is not.")
    
    if spike_cut_length < 0:
        raise ModelConfigurationException("Spike cut length must be non-negative.")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('preprocessor_values_path', help='path to preprocessor values json')
    parser.add_argument('method_config_path', help='path to method configuration json')
    parser.add_argument('output_path', help='path to store final model configuration')

    args = parser.parse_args()

    preprocessor_values = ju.read(args.preprocessor_values_path)
    method_config = ju.read(args.method_config_path)
    out_config = configure_model(method_config, preprocessor_values)

    ju.write(args.output_path, out_config)


if __name__ == "__main__": main()



import logging
import argparse
import subprocess
import os
import errno

import allensdk.core.json_utilities as ju
from allensdk.config.manifest import Manifest

SHARED_PYTHON = "/shared/utils.x86_64/python-2.7/bin/python"
SHARED_SDK = "/shared/bioapps/infoapps/lims2_modules/lib/allensdk"
RUN_PYTHON = "/shared/bioapps/infoapps/lims2_modules/lib/python/run_python.sh"

class PipelineModule( object ):
    def __init__(self, description="", parser=None):
        if parser is None:
            self.parser = default_argument_parser(description)
        else:
            self.parser = parser
            
        self._args = None

    @property
    def args(self):
        if self._args is None:
            self._args = self.parser.parse_args()
            logging.basicConfig(level=self.args.log_level, format="%(asctime)s:%(levelname)s:%(message)s")

        return self._args

    def input_data(self):
        try:
            return ju.read(self.args.input_json)
        except Exception as e:
            logging.error("could not read input json: %s", self.args.input_json)
            raise e
    
    def write_output_data(self, data):
        try:
            ju.write(self.args.output_json, data)
        except Exception as e:
            logging.error("could not write output json: %s", self.args.output_json)
            raise e 


def default_argument_parser(description=""):
    parser = argparse.ArgumentParser(description)
    parser.add_argument('input_json')
    parser.add_argument('output_json')
    parser.add_argument('--log-level', default=logging.DEBUG)

    return parser


def run_module(module, input_data, storage_directory, 
               optional_args=None, 
               python=SHARED_PYTHON,
               sdk_path=SHARED_SDK,
               local=False,
               pbs=None,
               conda_env=None):

    PBS_TEMPLATE="""
    export PYTHONPATH=%(sdk_path)s:$PYTHONPATH
    PYTHON=%(python)s
    SCRIPT="%(module)s"
    $PYTHON $SCRIPT %(optional_args)s %(input_json)s %(output_json)s 
    """

    if optional_args is None:
        optional_args = []

    input_json = os.path.join(storage_directory, "input.json")
    output_json = os.path.join(storage_directory, "output.json")
    pbs_file = os.path.join(storage_directory, "run.pbs")

    Manifest.safe_mkdir(storage_directory)

    pbs_headers = [ ('-j oe'),
                    ('-o %s' % os.path.join(storage_directory, "run.log")) ]
    pbs = pbs if pbs is not None else {}

    queue = pbs.get('queue', 'informatics')
    pbs_headers.append('-q %s' % queue)

    walltime = pbs.get('walltime', '3:00:00')
    pbs_headers.append('-l walltime=%s' % walltime)

    vmem = pbs.get('vmem', 16)
    pbs_headers.append('-l vmem=%dgb' % vmem)

    if 'job_name' in pbs:
        pbs_headers.append('-N %s' % pbs['job_name'])
       
    if 'ncpus' in pbs:
        pbs_headers.append('-l ncpus=%d' % pbs['ncpus'])

    pbs_headers = [ '#PBS %s' % s for s in pbs_headers ]

    if conda_env is not None:
        PBS_TEMPLATE = """
    module load anaconda/4.3.1
    source activate %(conda_env)s""" % {'conda_env':conda_env} + PBS_TEMPLATE

    with open(pbs_file,"w") as f:
        f.write('\n'.join(pbs_headers) + PBS_TEMPLATE % { 
                "python": python,
                "sdk_path": sdk_path,
                "module": module,
                "input_json": input_json,
                "output_json": output_json,
                "optional_args": " ".join(optional_args)
                })



    ju.write(input_json, input_data)

    if local:
        subprocess.call(['sh', pbs_file])
    else:
        subprocess.call(['qsub', pbs_file])
    

        
        
        
        


                                    



from PIL import Image, ImageDraw
import numpy as np

class MorphologyColors(object):
    def __init__(self):
        self.soma = (0, 0, 0)
        self.axon = (70, 130, 180)
        self.basal = (178, 34, 34)
        self.apical = (255, 127, 80)

    def set_soma_color(self, r, g, b):
        self.soma = (r, g, b)

    def set_axon_color(self, r, g, b):
        self.axon = (r, g, b)

    def set_basal_color(self, r, g, b):
        self.basal = (r, g, b)

    def set_apical_color(self, r, g, b):
        self.apical = (r, g, b)


# create empty image
def create_image(w, h, color=None, alpha=False):
    if alpha:
        mode = 'RGBA'
    else:
        mode = 'RGB'
    if color is not None:
        return Image.new(mode, (w,h), color)
    else:
        return Image.new(mode, (w,h))


def calculate_scale(morph, pix_width, pix_height):
    """ Calculates scaling factor and x,y insets required to auto-scale
        and center morphology into box with specified numbers of pixels

        Parameters
        ----------

        morph: AISDK Morphology object

        pix_width: int
        Number of image pixels on X axis

        pix_height: int
        Number of image pixels on Y axis

        Returns
        -------
        real, real, real
        First return value is the scaling factor. Second is the
        number of pixels needed to adjust x-coordinates so that the
        morphology is horizontally centered. Third is the number of 
        pixels needed to adjust the y-coordinates so that the morphology 
        is vertically centered.
    """
    dims, low, high = morph.get_dimensions()
    # get boundaries of morphology
    xlow = low[0]
    xhigh = high[0]
    ylow = low[1]
    yhigh = high[1]
    # determine scale on X and Y to make morphology fit in image area
    hscale = pix_width / (xhigh - xlow)
    vscale = pix_height / (yhigh - ylow)
    # select lowest scaling factor so morphology is stretched to
    #   maximum width/height along axis that is tightest fit
    #   and adjust inset on other axis so morphology is centered
    scale_factor = min(hscale, vscale)
    if hscale < vscale:
        # image constrained on horizontal axis
        scale_factor = hscale
        # center image vertically
        v_center = (ylow + yhigh) / 2.0
        scale_inset_x = -low[0] * scale_factor
        # invert y coordinates for conversion to pixel space
        scale_inset_y = pix_height/2 + scale_factor*v_center
    else:
        # image constrained on vertical axis
        scale_factor = vscale
        # center image horizontally
        h_center = (xlow + xhigh) / 2.0
        scale_inset_x = pix_width/2 - scale_factor*h_center
        scale_inset_y = -low[1] * scale_factor
    return scale_factor, scale_inset_x, scale_inset_y


# draw morphology on image -- takes image and morphology, modifies image
#       options: scale to fit | linear scaling
def draw_morphology(img, morph, 
        inset_left=0, inset_right=0, inset_top=0, inset_bottom=0, 
        scale_to_fit=False, scale_factor=1.0, colors=None):
    """ Draws morphology onto image
        When no scaling is applied, and no insets are provided, the 
        coordinates of the morphology are used directly -- i.e., 100 in
        morphology coordinates is equal to 100 pixels.

        The scale factor is multiplied to morphology coordinates before
        being drawn. If scale_factor=2 then 50 in morphology coordinates
        is 100 pixels. Left and top insets shift the coordinate axes
        for drawing. E.g., if left=10 and top=5 then 0,0 in morphology
        coordinates is 10,5 in pixel space. Bottom and right insets are
        ignored.
        
        If scale_to_fit is set then scale factor is ignored. The 
        morphology is scaled to be the maximum size that fits in
        the image, taking into account insets. In a 100x100 image, if
        all insets=10, then the image is scaled to fit into the center
        80x80 pixel area, and nothing is drawn in the inset border areas.

        Axons are drawn before soma and dendrite compartments.


        Parameters
        ----------
        
        img: PIL image object

        morph: AISDK Morphology object

        inset_*: real
        This is the number of pixels to use as border on top/bottom/
        right/left. If scale_to_fit is false then only the top/left
        values are used, as the scale_factor will determine how
        large the morphology is (it can be drawn beyond insets and even
        beyond image boundaries)

        scale_to_fit: boolean
        If true then morphology is scaled to the inset area of the 
        image and scale_factor is ignored. Morphology is centered
        in the image in the sense that the top/bottom and left/right
        edges of the morphology are equidistant from image borders.

        scale_factor: real
        A scalar amount that is multiplied to morphology coordinates
        before drawing

        colors: MorphologyColors object
        This is the color scheme used to draw the morphology. If 
        colors=None then default coloring is used

        Returns
        -------

        2-dimensional array, the pixel coordinates of the soma root [x,y]
    """
    # determine drawing area, scaling factor, offset to origin
    # if scaling to fit, find value that scales morphology so height
    #   or width matches image area. adjust scale_factor and insets
    #   as necessary so morphology can be drawn normally
    dims, low, high = morph.get_dimensions()
    if scale_to_fit:
        # get image area based on requested insets
        width, height = img.size
        pix_width = (width - inset_right) - inset_left
        pix_height = (height - inset_bottom) - inset_top
        # get scale and x,y insets from auto-scaling
        scale_factor, scale_inset_x, scale_inset_y = calculate_scale(morph, pix_width, pix_height)
    else:
        # no implicit inset necessary due to scaling
        scale_inset_x = 0
        scale_inset_y = 0

    # order compartments by depth to approximate 3D rendering
    sorted(morph.compartment_list, key=lambda x: x.node1.y)

    # if color not specified, select default
    if colors is None:
        colors = MorphologyColors()

    canvas = ImageDraw.Draw(img)

    for i in range(3):
        for comp in morph.compartment_list:
            if comp.node2.t == 1:
                if i != 2: # soma drawn last
                    # NOTE: there are unlikely to be soma compartments
                    # additional soma-drawing code is below
                    continue
                color = colors.soma
            elif comp.node2.t == 2:
                if i != 0: # axon drawn first
                    continue
                color = colors.axon
            elif comp.node2.t == 3:
                if i != 1: # dendrite drawn second
                    continue
                color = colors.basal
            elif comp.node2.t == 4:
                if i != 1: # dendrite drawn second
                    continue
                color = colors.apical
            x0 = scale_inset_x + inset_left + scale_factor * comp.node1.x
            x1 = scale_inset_x + inset_left + scale_factor * comp.node2.x
            # y coordinate inverted because morphology values are
            #   increasing going 'up while pixel values increase
            #   going down
            y0 = scale_inset_y + inset_top - scale_factor * comp.node1.y
            y1 = scale_inset_y + inset_top - scale_factor * comp.node2.y
            canvas.line((x0, y0, x1, y1), color)

    # a compartment type is defined by the type of the 2nd node defining
    #   the compartment. if there is a single soma node, there can be
    #   no soma compartments. draw the root soma node
    root = morph.soma_root()
    x = scale_inset_x + inset_left + scale_factor * root.x
    y = scale_inset_y + inset_top - scale_factor * root.y
    rad = scale_factor * root.radius
    x0 = int(x - rad)
    y0 = int(y - rad)
    x1 = int(x0 + 2*rad)
    y1 = int(y0 + 2*rad)
    canvas.ellipse((x0,y0,x1,y1), fill=colors.soma, outline=colors.soma)

    # return soma root coordinate, in unit of pixels
    return [x, y]


def draw_density_hist(img, morph, vert_scale,
        inset_left=0, inset_right=0, inset_top=0, inset_bottom=0, 
        num_bins=None, colors=None):
    """ Draws density histogram onto image
        When no scaling is applied, and no insets are provided, the 
        coordinates of the morphology are used directly -- i.e., 100 in
        morphology coordinates is equal to 100 pixels.

        The scale factor is multiplied to morphology coordinates before
        being drawn. If scale_factor=2 then 50 in morphology coordinates
        is 100 pixels. Left and top insets shift the coordinate axes
        for drawing. E.g., if left=10 and top=5 then 0,0 in morphology
        coordinates is 10,5 in pixel space. Bottom and right insets are
        ignored.
        
        If scale_to_fit is set then scale factor is ignored. The 
        morphology is scaled to be the maximum size that fits in
        the image, taking into account insets. In a 100x100 image, if
        all insets=10, then the image is scaled to fit into the center
        80x80 pixel area, and nothing is drawn in the inset border areas.

        Axons are drawn before soma and dendrite compartments.


        Parameters
        ----------
        
        img: PIL image object

        morph: AISDK Morphology object

        vert_scale: real
        This is the amout required to multiply to a moprhology
        y-coordinate to convert it to relative cortical depth (on [0,1]).
        This is the inverse of the cortical thickness.

        inset_*: real
        This is the number of pixels to use as border on top/bottom/
        right/left. If scale_to_fit is false then only the top/left
        values are used, as the scale_factor will determine how
        large the morphology is (it can be drawn beyond insets and even
        beyond image boundaries)

        num_bins: int
        The number of bins in the histogram

        colors: MorphologyColors object
        This is the color scheme used to draw the morphology. If 
        colors=None then default coloring is used

        Returns
        -------

        Histogram arrays: [hist, hist2, hist3, hist4]
        where hist is the histgram of all neurites, and hist[234] are
        the histograms of SWC types 2,3,4
    """
    # if number of bins not specified, default to vertical size of 
    #   drawing area in image
    img_width, img_height = img.size
    draw_width = img_width - (inset_left + inset_right)
    draw_height = img_height - (inset_top + inset_bottom)
    if num_bins is None:
        num_bins = draw_height
    # histograms for each analyzed SWC type
    hist_2 = np.zeros(num_bins)
    hist_3 = np.zeros(num_bins)
    hist_4 = np.zeros(num_bins)
    # total response in a bin
    hist = np.zeros(num_bins)

    print "Vert scale", vert_scale

    # if color not specified, select default
    if colors is None:
        colors = MorphologyColors()

    canvas = ImageDraw.Draw(img)

    # for each compartment, split its length (weight) between bins for 
    #   start and end nodes
    for seg in morph.compartment_list:
        wt = seg.length / 2.0
        # node 1
        bin1 = int(-vert_scale * seg.node1.y * num_bins)
        if bin1 == num_bins:
            bin1 = num_bins-1
        # only include parts of the histogram that are in the viewable
        #   range (ie, exclude parts that extend to pia and/or wm)
        if bin1 >= 0 and bin1 < num_bins:
            if seg.node1.t == 2:
                hist_2[bin1] += wt
            elif seg.node1.t == 3:
                hist_3[bin1] += wt
            elif seg.node1.t == 4:
                hist_4[bin1] += wt
            hist[bin1] += wt
        # node 2
        bin2 = int(-vert_scale * seg.node1.y * num_bins)
        if bin2 == num_bins:
            bin2 = num_bins-1
        if bin2 >= 0 and bin2 < num_bins:
            if seg.node2.t == 2:
                hist_2[bin2] += wt
            elif seg.node2.t == 3:
                hist_3[bin2] += wt
            elif seg.node2.t == 4:
                hist_4[bin2] += wt
            hist[bin2] += wt

    # draw axis line
    col = (128, 128, 128, 256)
    x0 = inset_left
    x1 = x0
    y0 = inset_top
    y1 = img_height-inset_bottom
    canvas.line((x0, y0, x1, y1), col)

    hist_step = 1.0 * draw_height / num_bins;
    hist_scale = (draw_width-1) / hist.max()
    for seg in morph.compartment_list:
        ypos = 1.0 * inset_top
        for i in range(num_bins):
            y0 = int(ypos)
            y1 = int(ypos + hist_step)
            x0 = int(1 + inset_left)
            x1 = int(1 + inset_left + hist_2[i] * hist_scale + 0.99)
            if x0 != x1:
                ytmp = ypos
                while ytmp < y1:
                    canvas.line((x0, int(ytmp), x1, int(ytmp)), colors.axon)
                    ytmp += hist_step
            x0 = x1
            x1 += int(hist_3[i] * hist_scale + 0.99)
            if x0 != x1:
                ytmp = ypos
                while ytmp < y1:
                    canvas.line((x0, int(ytmp), x1, int(ytmp)), colors.basal)
                    ytmp += hist_step
            x0 = x1
            x1 += int(hist_4[i] * hist_scale + 0.99)
            if x0 < x1:
                ytmp = ypos
                while ytmp < y1:
                    canvas.line((x0, int(ytmp), x1, int(ytmp)), colors.apical)
                    ytmp += hist_step
            ypos += hist_step

    return hist, hist_2, hist_3, hist_4

# TODO
# draw path on image -- takes image and path, modifies image
#def draw_path(img, path, color):
    # path is in units of pixels
    # foreach vertex, draw line in specified color

# refine section boundaries -- takes labeled regions and generates labeled mask
# -> need constraints on input. lookup of points is easy. categorizing
#   what pixel is part of what ask is not without expanding masks laterally

# label morphology -- takes morpholgoy and labeled mask and adds lables
#   to each compartment





import numpy as np

"""
TODO: license
TODO: comment style
"""

def AIC(RSS, k, n):
    """
    Computes the Akaike Information Criterion.
 
       RSS-residual sum of squares of the fitting errors.
       k  - number of fitted parameters.
       n  - number of observations.
    """
    AIC = 2 * k + n * np.log( RSS/n)
    return AIC
 
def AICc(RSS, k, n):
    """
    Corrected AIC. formula from Wikipedia.
    """
    retval = AIC(RSS, k, n)
    if  n-k-1 != 0:
        retval += 2.0 *k* (k+1)/ (n-k-1)
    return retval
 
def BIC(RSS, k, n):
    """
    Bayesian information criterion or Schwartz information criterion.
    Formula from wikipedia.
    """
    return n * np.log(RSS/n) + k * np.log(n)



import matplotlib
matplotlib.use('agg')

import logging
import numpy as np
import os, sys

from allensdk.internal.core.lims_pipeline_module import PipelineModule, run_module
from allensdk.internal.brain_observatory.run_itracker import (run_itracker, 
                                                              compute_bounding_box, 
                                                              DEFAULT_THRESHOLD_FACTOR,
                                                              get_experiment_info)

def debug(experiment_id, num_frames=None, threshold_factor=None, local=False):
    OUTPUT_DIR = "/data/informatics/CAM/eye_tracking/"
    SDK_PATH = "/data/informatics/CAM/eye_tracking/allensdk"
    SCRIPT_PATH = "/data/informatics/CAM/eye_tracking/allensdk/allensdk/internal/pipeline_modules/run_eye_tracking.py" 
    
    experiment_dir = os.path.join(OUTPUT_DIR, str(experiment_id))
    
    info = get_experiment_info(experiment_id)
    info['output_directory'] = experiment_dir

    optional_args = [ ]
    if num_frames is not None:
        optional_args += ['--num_frames',str(num_frames)]

    run_module(SCRIPT_PATH, 
               info, 
               experiment_dir, 
               sdk_path=SDK_PATH,
               pbs=dict(vmem=160,
                        job_name="itrack_%d"% experiment_id,
                        walltime="10:00:00"),
               local=local,
               optional_args=optional_args)
    
def main():
    mod = PipelineModule()
    mod.parser.add_argument("--num_frames", type=int, default=None)
    mod.parser.add_argument("--threshold_factor", type=float, default=DEFAULT_THRESHOLD_FACTOR)

    data = mod.input_data()
    args = dict(
        movie_file=data['movie_file'],
        metadata_file=data['metadata_file'],
        output_directory=data['output_directory'],
        threshold_factor=data.get('threshold_factor', mod.args.threshold_factor),
        num_frames=mod.args.num_frames,
        auto=True,
        cache_input_frames=True,
        input_block_size=None,
        output_annotated_movie_block_size=None
        )

    if data.get('pupil_points', None):
        args['bbox_pupil'] = compute_bounding_box(data['pupil_points'])
    if data.get('corneal_reflection_points', None):
        args['bbox_cr'] = compute_bounding_box(data['corneal_reflection_points'])
    
    tracker = run_itracker(**args)

    logging.debug("finished running itracker")
    
    output_data = dict(
      pupil_file=tracker.pupil_file,
      corneal_reflection_file=tracker.cr_file,
      mean_frame_file=tracker.mean_frame_file
      )
    
    mod.write_output_data(output_data)
  
if __name__=='__main__': main()



import numpy as np
from scipy.signal import correlate2d
from scipy.signal import fftconvolve
from scipy.ndimage.filters import sobel
import logging

def default_ray(n):

    y = np.zeros(n,dtype=np.int64)
    x = np.arange(n,dtype=np.int64)

    return np.vstack([y,x])

def rotate_ray(ray,theta):

    y,x = ray.astype(np.float64)

    xp = x*np.cos(theta) + y*np.sin(theta)
    yp = -x*np.sin(theta) + y*np.cos(theta)

    return np.vstack([yp.astype(np.int64),xp.astype(np.int64)])

def generate_rays(image_array, seed_pixel):

    N = 18 #200

    #mag, grad_x, grad_y = sobel_grad(image_array.astype('float'))

    shape = image_array.shape
    Y,X = np.mgrid[:shape[1],:shape[0]]

    n = int(np.sqrt(shape[0]**2 + shape[1]**2))
    angles = np.arange(N)*2.0*np.pi/N
    rays = []

    tangents = []

    ray_grads = []

    good_coords_mask = lambda y,x:  np.logical_and(np.logical_and(y>=0,y<shape[0]),np.logical_and(x>=0,x<shape[1]))

    for theta in angles:
        new_ray = rotate_ray(default_ray(n),theta)
        new_ray = new_ray.T + seed_pixel
        new_ray = new_ray.T


        mask = good_coords_mask(new_ray[0],new_ray[1])

        ym = new_ray[0][mask]
        xm = new_ray[1][mask]

        rays += [np.vstack([ym,xm])]

        t = np.array([np.sin(theta),np.cos(theta)])
        tangents += [t]

        #rg = t[1]*grad_x[ym,xm] + t[0]*grad_y[ym,xm]
        #rg[rg<0] = 0.0
        rg = image_array[ym,xm]
        #rg = rg[1:].astype(np.float64) - rg[:-1].astype(np.float64)
        # rg[rg<0] = 0.0
        ray_grads += [rg]


    return rays, ray_grads

def initial_pupil_point(image_array, bbox=None):
    """bbox is a tuple of (xmin, xmax, ymin, ymax)"""

    if bbox is not None:
        xmin, xmax, ymin, ymax = bbox
        crop_im = image_array[ymin:ymax, xmin:xmax]
    else:
        shape = image_array.shape
        crop_distance = 50
        crop_im = image_array[crop_distance:shape[0]-crop_distance, crop_distance:shape[1]-crop_distance]

    m = np.max(crop_im)

    dark_square = m*np.ones([30,30])
    #c = correlate2d(m-crop_im,dark_square,mode='same')
    c = fftconvolve(m-crop_im,dark_square[::-1,::-1],mode='same')
    y,x = np.where(c==np.max(c))

    if bbox is not None:
        ybar=int(np.mean(y))+ymin
        xbar=int(np.mean(x))+xmin
    else:
        ybar=int(np.mean(y))+crop_distance
        xbar=int(np.mean(x))+crop_distance

    return ybar, xbar

def initial_cr_point(image_array, bbox=None):
    """bbox is a tuple of (xmin, xmax, ymin, ymax)"""

    if bbox is not None:
        xmin, xmax, ymin, ymax = bbox
        crop_im = image_array[ymin:ymax, xmin:xmax]
    else:
        shape = image_array.shape
        crop_distance = 50
        crop_im = image_array[crop_distance:shape[0]-crop_distance, crop_distance:shape[1]-crop_distance]

    m = np.max(crop_im)
    mean = np.mean(crop_im)

    Y,X = np.meshgrid(np.arange(-20,20),np.arange(-20,20))
    bright_circle = np.zeros([40,40])
    mask = X**2 + Y**2 < 100.
    bright_circle[mask] = m
    bright_circle -= np.mean(bright_circle)

    #c = correlate2d(crop_im-mean,bright_circle,mode='same')
    c = fftconvolve(crop_im-mean,bright_circle[::-1,::-1],mode='same')
    y,x = np.where(c==np.max(c))

    if bbox is not None:
        ybar=int(np.mean(y))+ymin
        xbar=int(np.mean(x))+xmin
    else:
        ybar=int(np.mean(y))+crop_distance
        xbar=int(np.mean(x))+crop_distance

    return ybar, xbar

def sobel_grad(image_array):

    grad_y = sobel(image_array.astype(np.float64),0)
    grad_x = sobel(image_array.astype(np.float64),1)

    #print "grad_x dtype = ", grad_x.dtype

    mag = np.sqrt(grad_y**2 + grad_x**2) + 1e-16

    return mag, grad_x, grad_y

def medfilt_custom(x, kernel_size=3):
    '''This median filter returns 'nan' whenever any value in the kernal width is 'nan' and the median otherwise'''
    T = x.shape[0]
    delta = kernel_size/2

    x_med = np.zeros(x.shape)
    window = x[0:delta+1]
    if np.any(np.isnan(window)):
        x_med[0] = np.nan
    else:
        x_med[0] = np.median(window)

    # print window
    for t in range(1,T):
        window = x[t-delta:t+delta+1]
        # print window
        if np.any(np.isnan(window)):
            x_med[t] = np.nan
        else:
            x_med[t] = np.median(window)

    return x_med

def eccentricity(a1, a2):

    return np.sqrt(1.0 - (np.minimum(a1,a2)**2)/(np.maximum(a1,a2)**2))

def median_absolute_deviation(a, consistency_constant=1.4826):
    '''Calculate the median absolute deviation of a univariate dataset.

    Parameters
    ----------
    a : numpy.ndarray
        Sample data.
    consistency_constant : float
        Constant to make the MAD a consistent estimator of the population
        standard deviation (1.4826 for a normal distribution).

    Returns
    -------
    float
        Median absolute deviation of the data.
    '''
    return consistency_constant * np.nanmedian(np.abs(a - np.nanmedian(a)))

def post_process_cr(cr_params):
    """This will replace questionable values of the CR x and y position with 'nan'

        1)  threshold ellipse area by 99th percentile area distribution
        2)  median filter using custom median filter
        3)  remove deviations from discontinuous jumps

        The 'nan' values likely represent obscured CRs, secondary reflections, merges
        with the secondary reflection, or visual distortions due to the whisker or
        deformations of the eye"""

    area = np.pi*cr_params.T[3]*cr_params.T[4]

    # compute a threshold on the area of the cr ellipse
    dev = median_absolute_deviation(area)
    if dev == 0:
        logging.warning("Median absolute deviation is 0,"
                        "falling back to standard deviation.")
        dev = np.nanstd(area)
    threshold = np.nanmedian(area) + 3*dev

    x_center = cr_params.T[0]
    y_center = cr_params.T[1]

    # set x,y where area is over threshold to nan
    x_center[area>threshold] = np.nan
    y_center[area>threshold] = np.nan

    # median filter
    x_center_med = medfilt_custom(x_center, kernel_size=3)
    y_center_med = medfilt_custom(y_center, kernel_size=3)

    x_mask_finite = np.where(np.isfinite(x_center_med))[0]
    y_mask_finite = np.where(np.isfinite(y_center_med))[0]

    # if y increases discontinuously or x decreases discontinuously,
    #  that is probably a CR secondary reflection
    mean_x = np.mean(x_center_med[x_mask_finite])
    mean_y = np.mean(y_center_med[y_mask_finite])

    std_x = np.std(x_center_med[x_mask_finite])
    std_y = np.std(y_center_med[y_mask_finite])

    # set these extreme values to nan
    #x_center_med[x_center_med < mean_x - 3*std_x] = np.nan
    #y_center_med[y_center_med > mean_y + 3*std_y] = np.nan
    x_center_med[np.abs(x_center_med - mean_x) > 3*std_x] = np.nan
    y_center_med[np.abs(y_center_med - mean_y) > 3*std_y] = np.nan

    either_nan_mask = np.logical_and(np.isnan(x_center_med),np.isnan(y_center_med))
    x_center_med[either_nan_mask] = np.nan
    y_center_med[either_nan_mask] = np.nan

    new_cr = np.vstack([x_center_med, y_center_med]).T

    bad_points_mask = either_nan_mask

    return new_cr, bad_points_mask


def post_process_pupil(pupil_params):
    '''Filter pupil parameters to replace outliers with nan

    Parameters
    ----------
    pupil_params : numpy.ndarray
        (Nx5) array of pupil parameters [x, y, angle, axis1, axis2].

    Returns
    -------
    numpy.ndarray
        Pupil parameters with outliers replaced with nan
    '''
    area = np.pi*pupil_params.T[3]*pupil_params.T[4]
    threshold = np.percentile(area[np.isfinite(area)], 99)
    outlier_index = area > threshold
    pupil_params[outlier_index, :] = np.nan
    return pupil_params


def filter_bad_params(params, frame_width, frame_height):
    '''Replace positions outside image with nan'''
    params[(params[:,0] > frame_width) | (params[:,0] < 0), :] = np.nan
    params[(params[:,1] > frame_height) | (params[:,1] < 0), :] = np.nan
    return params



import argparse, logging
import itertools
from scipy.optimize import fmin
import numpy as np
import os
import allensdk.core.json_utilities as ju
import allensdk.internal.model.glif.find_sweeps as fs
from allensdk.internal.model.data_access import load_sweeps
from allensdk.internal.model.glif.MLIN import MLIN
from allensdk.internal.model.glif.ASGLM import ASGLM_pairwise
from allensdk.internal.model.glif.rc import least_squares_RCEl_calc_tested
from allensdk.internal.model.glif.threshold_adaptation import calc_spike_component_of_threshold_from_multiblip
from allensdk.internal.model.glif.spike_cutting import calc_spike_cut_and_v_reset_via_expvar_residuals
from allensdk.internal.model.glif.find_spikes import find_spikes_list, find_spikes_ssq_list
from allensdk.internal.model.glif.threshold_adaptation import fit_avoltage_bvoltage_th, fit_avoltage_bvoltage
import allensdk.ephys.ephys_extractor as efex
import allensdk.ephys.ephys_features as ft
from allensdk.model.glif.glif_neuron_methods import spike_component_of_threshold_exact
import matplotlib.pyplot as plt
import allensdk.internal.model.glif.plotting as plotting

RESTING_POTENTIAL = 'slow_vm_mv'
DEFAULT_DT = 5e-05
DEFAULT_CUT = 0
DEFAULT_BESSEL = { 'N': 4, 'freq': 10000 }
MAKE_PLOT = True
SHOW_PLOT = False
SAVE_FIG =True
SHORT_RUN = False

class MissingSpikeException(Exception): pass

RESTING_POTENTIAL = 'slow_vm_mv'

def find_first_spike_voltage(voltage, dt, ssq=False, MAKE_PLOT=False, SHOW_PLOT=False, BLOCK=False,
                             dv_cutoff=20.0, thresh_frac=0.05):
    '''calculate voltage at threshold of first spike
    Parameters
    ----------
    voltage: numpy array
        voltage trace
    dt: float
        sampling time step
    ssq: Boolean
        whether there is or is not a subrathreshold short square pulse (note that if thes
    MAKE_PLOT: Boolean
        specifies whether or not a plot should be made
    SHOW_PLOT: Boolean
        specifies if a visualization should be made
    BLOCK: Boolean
        if a plot is made this specifies weather to stop the code until the plot is closed
    dv_cutoff: float
        specifies cut off of the derivative of the voltage
    thresh_frac: float
        variable that goes into feature extractor
    
    Returns
    -------
    :float
        voltage of threshold of first spike
    '''
    
    if ssq:
        spike_time_steps, _ = find_spikes_ssq_list([voltage], dt, dv_cutoff=dv_cutoff, thresh_frac=thresh_frac)
    else:
        spike_time_steps, _ = find_spikes_list([voltage], dt)

    if MAKE_PLOT:
        plotting.plotSpikes([voltage], spike_time_steps, dt, blockME=False, method='dvdt_v2')
    if SHOW_PLOT:
        plt.show(block=BLOCK)

    if len(spike_time_steps[0]) == 0:
        raise MissingSpikeException('No spike detected.')
    
    return voltage[spike_time_steps[0][0]]

def tag_plot(tag, fs=9):
    plt.annotate(tag, xy=(0.98, .01),
                 xycoords='figure fraction',
                 horizontalalignment='right', 
                 verticalalignment='bottom',
                 fontsize=fs)
    
def estimate_dv_cutoff(voltage_list, dt, start_t, end_t):
    v_set = [ v * 1e3 for v in voltage_list ]
    t_set = [ np.arange(0, len(v)) * dt for v in voltage_list ]
    
    dv_cutoff, thresh_frac = ft.estimate_adjusted_detection_parameters(v_set, t_set,
                                                                       start_t, end_t,
                                                                       filter=None)
    
    return dv_cutoff, thresh_frac

def preprocess_neuron(nwb_file, sweep_list, cell_properties=None,
                      dt=None, cut=None, bessel=None, save_figure_path=None):
    if dt is None:
        dt = DEFAULT_DT  
    if cut is None:
        cut = DEFAULT_CUT  
    if bessel is None:
        bessel = DEFAULT_BESSEL

    sweep_index = { s['sweep_number']: s for s in sweep_list }

    noise_sweeps = fs.find_noise_sweeps(sweep_index)
    noise1_sweeps = noise_sweeps['noise1']
    noise2_sweeps = noise_sweeps['noise2']

    ssq_sweeps = fs.find_short_square_sweeps(sweep_index) 
    all_ssq_data = load_sweeps(nwb_file, ssq_sweeps['all'], dt, cut, bessel)
    ssq_dv_cutoff, ssq_thresh_frac = estimate_dv_cutoff(all_ssq_data['voltage'], dt, 
                                                        efex.SHORT_SQUARES_WINDOW_START,
                                                        efex.SHORT_SQUARES_WINDOW_END)
    
    ssq_triple_sweeps = ssq_sweeps['triple']
    
    ramp_sweeps = fs.find_ramp_sweeps(sweep_index)['suprathreshold']
    R2R_sweeps = fs.find_ramp_to_rheo_sweeps(sweep_index)['all']
    
    noise1_data = load_sweeps(nwb_file, noise1_sweeps, dt, cut, bessel)
    noise2_data = load_sweeps(nwb_file, noise2_sweeps, dt, cut, bessel)

    maximum_subthreshold_short_square_sweeps = ssq_sweeps['maximum_subthreshold']
    maximum_subthreshold_short_square_data = load_sweeps(nwb_file, [maximum_subthreshold_short_square_sweeps[0]], dt, cut, bessel)
    minimum_suprathreshold_short_square_sweeps = ssq_sweeps['minimum_suprathreshold']
    minimum_suprathreshold_short_square_data = load_sweeps(nwb_file, [minimum_suprathreshold_short_square_sweeps[0]], dt, cut, bessel)

    dt = noise1_data['dt'][0]  #getting subsampled dt returned for ease of use

    subthresh_noise_current_list=[]
    subthresh_noise_voltage_list=[]
    noise_El_list=[]
    for ss in range(0, len(noise1_data['current'])):
        #--subthreshold noise has first epoch of noise with a region of no stimulation before and after (note the selection of end point is hard coded)
        subthresh_noise_current_list.append(noise1_data['current'][ss][noise1_data['start_idx'][ss]:int(6./dt)])
        subthresh_noise_voltage_list.append(noise1_data['voltage'][ss][noise1_data['start_idx'][ss]:int(6./dt)])
        noise_El_list.append(sweep_index[noise1_sweeps[ss]][RESTING_POTENTIAL]*1e-3)  

    # Els calculated from QC
    El_noise=np.mean(noise_El_list)
    El_subthreshold_blip=sweep_index[maximum_subthreshold_short_square_sweeps[0]][RESTING_POTENTIAL]*1e-3
    El_suprathreshold_blip=sweep_index[minimum_suprathreshold_short_square_sweeps[0]][RESTING_POTENTIAL]*1e-3

    if len(ramp_sweeps):
        logging.info('has ramp')
        ramp_data = load_sweeps(nwb_file, ramp_sweeps, dt, cut, bessel)
        El_ramp=sweep_index[ramp_sweeps[0]][RESTING_POTENTIAL]*1e-3
    else:
        ramp_sweeps=None
        ramp_data=None
        El_ramp=None
        logging.info("No ramp")

    if len(ssq_triple_sweeps):
        logging.info('has multi ss')
        multi_ssq_data = load_sweeps(nwb_file, ssq_triple_sweeps, dt, cut, bessel)
        El_multi_ssq_data=sweep_index[ssq_triple_sweeps[0]][RESTING_POTENTIAL]*1e-3
        multi_ssq_dv_cutoff, multi_ssq_thresh_frac = estimate_dv_cutoff(multi_ssq_data['voltage'], dt, 
                                                                        efex.SHORT_SQUARE_TRIPLE_WINDOW_START,
                                                                        efex.SHORT_SQUARE_TRIPLE_WINDOW_END)
        print "*************************"
        print "ssq", ssq_dv_cutoff, ssq_thresh_frac
        print "triple",multi_ssq_dv_cutoff, multi_ssq_thresh_frac
    else:
        ssq_triple_sweeps=None
        multi_ssq_data = None
        El_multi_ssq_data = None
        logging.info("No multi short square")

    # Needed for MLIN
    long_square_config = fs.find_long_square_sweeps(sweep_index)
    long_square_sweeps = long_square_config['all']
    subthreshold_long_square_sweeps = long_square_config['subthreshold']
    maximum_subthreshold_long_square_sweeps = long_square_config['maximum_subthreshold']
    #TODO: Here you are loading just one sweep: probably should load all
    maximum_subthreshold_long_square_data = load_sweeps(nwb_file, [maximum_subthreshold_long_square_sweeps[0]], dt, cut, bessel)
    El_max_subth_long_square=sweep_index[maximum_subthreshold_long_square_sweeps[0]][RESTING_POTENTIAL]*1e-3  

    #---------------------------------------------------------------        
    #---------find spiking indicies of spikes in noise--------------
    #---------------------------------------------------------------

    # note that when using find_spikes_list without removing the testpulse a warning will result from calculating feature_data['base_v'] in the feature extractor (line 375) this not relavent here
    noise1_ind_wo_test_pulse_removed, _ = find_spikes_list(noise1_data['voltage'], dt)
    noise2_ind_wo_test_pulse_removed, _ = find_spikes_list(noise2_data['voltage'], dt)
    #Put all ISI ind in a 
    ISI_length=np.array([])
    for ii in range(len(noise1_ind_wo_test_pulse_removed)):
        ISI_length=np.append(ISI_length,noise1_ind_wo_test_pulse_removed[ii][1:]-noise1_ind_wo_test_pulse_removed[ii][:-1])
    for ii in range(len(noise2_ind_wo_test_pulse_removed)):
        ISI_length=np.append(ISI_length,noise2_ind_wo_test_pulse_removed[ii][1:]-noise2_ind_wo_test_pulse_removed[ii][:-1])
    min_ISI_len=np.min(ISI_length)


    #-------------------------------------------------------------------------------------------------------------------
    #---------------------Compute R, C and EL via least squares-------------------------------------------------
    #------------------------------------------------------------------------------------------------------------------- 

    #--compute R, C, and El via least squares tested in verify_RCEl_GLM_vs_lssq_and_smooth.py
    (R_test_list, C_test_list, El_test_list)=least_squares_RCEl_calc_tested(subthresh_noise_voltage_list, subthresh_noise_current_list, dt)
    R_test_list_mean=np.mean(R_test_list)
    C_test_list_mean=np.mean(C_test_list)
    El_test_list_mean=np.mean(El_test_list)
    
    #-----------------------------------------------------------------------------------------        
    #------------------------ compute spike cut length----------------------------------------
    #-----------------------------------------------------------------------------------------

    #TODO: I should disentangle this function so I can get rid of the deltaV dependency 
    (spike_cut_length_NODELTAV, slope_at_min_expVar_list_NODELTAV, intercept_at_min_expVar_list_NODELTAV) \
        = calc_spike_cut_and_v_reset_via_expvar_residuals(noise1_data['current'], noise1_data['voltage'], 
                                                          dt, El_noise, 0, 
                                                          max_spike_cut_time=min_ISI_len*dt,
                                                          MAKE_PLOT=MAKE_PLOT, 
                                                          SHOW_PLOT=SHOW_PLOT, 
                                                          BLOCK=False)
    if SAVE_FIG:
        tag='spikeCutting_noDeltaV_regression.png'
        tag_plot(tag)
        plt.savefig(os.path.join(save_figure_path,tag), format='png')
        plt.close()
        tag='spikeCutting_noDeltaV_spike_wave_form.png'
        tag_plot(tag)
        plt.savefig(os.path.join(save_figure_path,tag), format='png')
        plt.close()
    logging.info('spike cut length: %d', spike_cut_length_NODELTAV)

    #-----------------------------------------------------------------------------------------        
    #------------------------ compute ASC amplitudes------------------------------------------
    #-----------------------------------------------------------------------------------------

    #***Hack: k's are being hard coded into this function and are not necessarily consistent with what is in the setting of AS currents!!!
    k_asc_possible=np.array([3, 10., 30., 100., 300.]) 
    if SHORT_RUN:  #THIS IS JUST FOR DEBUGGING SO THAT YOU DONT HAVE TO WAIT FOR THE ENTIRE MODULE TO RUN
        (best_k_pair_fit_ascR, best_asc_amp_fit_ascR, best_R_fit_ascR, best_llh_fit_ascR)=ASGLM_pairwise(k_asc_possible, 
                            noise1_data['current'], noise1_data['voltage'], noise1_ind_wo_test_pulse_removed, 
                            C_test_list_mean, C_test_list_mean*R_test_list_mean, spike_cut_length_NODELTAV, dt, El_noise,
                            SHORT_RUN=True, MAKE_PLOT=MAKE_PLOT, SHOW_PLOT=SHOW_PLOT, BLOCK=False)
        asc_amp_from_ASGLM=np.mean(best_asc_amp_fit_ascR, axis=0)
        R_from_ASGLM=np.mean(best_R_fit_ascR)

    else:
        (best_k_pair_fit_ascR, best_asc_amp_fit_ascR, best_R_fit_ascR, best_llh_fit_ascR)=ASGLM_pairwise(k_asc_possible, 
                            noise1_data['current'], noise1_data['voltage'], noise1_ind_wo_test_pulse_removed, 
                            C_test_list_mean, C_test_list_mean*R_test_list_mean, spike_cut_length_NODELTAV, dt, El_noise,
                            SHORT_RUN=False, MAKE_PLOT=MAKE_PLOT, SHOW_PLOT=SHOW_PLOT, BLOCK=False)            
        asc_amp_from_ASGLM=np.mean(best_asc_amp_fit_ascR, axis=0)
        R_from_ASGLM=np.mean(best_R_fit_ascR) 

    if SAVE_FIG:
        tag='GLM_fit_ascR_basis.png'
        tag_plot(tag)
        plt.savefig(os.path.join(save_figure_path,tag), format='png')
        plt.close()
        tag='GLM_fit_ascR_sumASC.png'
        tag_plot(tag)
        plt.savefig(os.path.join(save_figure_path,tag), format='png')
        plt.close()
        tag='GLM_fit_ascR_individualASC.png'
        tag_plot(tag)
        plt.savefig(os.path.join(save_figure_path,tag), format='png')
        plt.close()

    logging.info('Output of ASC fitting GLM')
    logging.info('R out_of_GLM_Rfit_Cfixed %f %s', R_from_ASGLM/1e6, "MOhms")
    logging.info('ASC amplitudes at the time of cut spike %s', str(asc_amp_from_ASGLM*1e12))
    logging.info("ks used %s", str(best_k_pair_fit_ascR))

    #-----------------------------------------------------------------------------------------        
    #------------------------ calculate thresholds-----------------------------------------
    #-----------------------------------------------------------------------------------------

    # ---extract instantaneous threshold from suprathreshold blip 
    try:
        th_inf_via_Vmeasure = find_first_spike_voltage(minimum_suprathreshold_short_square_data['voltage'][0][minimum_suprathreshold_short_square_data['start_idx'][0]:], 
                                                           dt,         
                                                           ssq=True,                                    
                                                           MAKE_PLOT=MAKE_PLOT, 
                                                           SHOW_PLOT=SHOW_PLOT, 
                                                           BLOCK=False,
                                                           dv_cutoff=ssq_dv_cutoff,
                                                           thresh_frac=ssq_thresh_frac)
        th_inf_via_Vmeasure_from0=th_inf_via_Vmeasure-El_suprathreshold_blip
        if SAVE_FIG:
            tag='th_inf_from_blip.png'
            tag_plot(tag)
            plt.savefig(os.path.join(save_figure_path, tag), format='png')
            plt.close()
    except MissingSpikeException as e:
        raise MissingSpikeException("The suprathreshold short square sweep must have a spike, but no spike was detected. This means that feature extraction and GLIF spike detection are inconsistent.")

    #-----------------------------------------------------------------------------------------------------
    #-----------------find spike and voltage component of the threshold---------------------------
    #-----------------------------------------------------------------------------------------------------

    # If a multishort square stimulus exists calculate spike component of threshold..
    if multi_ssq_data:
        (a_spike_component_of_threshold, b_spike_component_of_threshold, 
         mean_voltage_first_spike_of_blip) = calc_spike_component_of_threshold_from_multiblip(multi_ssq_data, 
                                                                     dt, 
                                                                     multi_ssq_dv_cutoff,
                                                                     multi_ssq_thresh_frac,
                                                                     MAKE_PLOT=MAKE_PLOT, 
                                                                     SHOW_PLOT=False, 
                                                                     BLOCK=False,
                                                                     PUBLICATION_PLOT=False)
        #adjust values to be after spike cutting
        if a_spike_component_of_threshold is not None and b_spike_component_of_threshold is not None:
            a_spike_component_of_threshold=spike_component_of_threshold_exact(a_spike_component_of_threshold, b_spike_component_of_threshold, spike_cut_length_NODELTAV*dt)

        if SAVE_FIG:
            tag='multiblip_fit.png'
            tag_plot(tag)
            plt.savefig(os.path.join(save_figure_path, tag), format='png')
            plt.close() 

            tag='multiblip_data.png'
            tag_plot(tag, fs=9)
            plt.savefig(os.path.join(save_figure_path,tag), format='png')
            plt.close() 

        #---calculate voltage componet of threshold 
        if a_spike_component_of_threshold is None or b_spike_component_of_threshold is None:
            logging.warning("spike component of threshold could not be calculated from the multiblip data")
            a_voltage_comp_of_thr_from_fitab=None
            b_voltage_comp_of_thr_from_fitab=None
            a_voltage_comp_of_thr_from_fitabth=None
            b_voltage_comp_of_thr_from_fitabth=None
            th_inf_fit_w_v_comp_of_th=None
            th_inf_fit_w_v_comp_of_th_from0=None
        else:
            #TODO:this function needs to be changed to use experimental change of reference
            b_voltage_guess = 5.0         
            a_voltage_guess = 0.1*b_voltage_guess
            fit_ab_vcomp_from_noise = fmin(func=fit_avoltage_bvoltage, 
                                           args=(noise1_data['voltage'], 
                                                 noise_El_list, 
                                                 spike_cut_length_NODELTAV, 
                                                 noise1_ind_wo_test_pulse_removed, #NOTE THAT IF YOU WANT TO USE THIS TO GET A VOLTAGE WITHIN THE FUNCTION YOU NEED TO SUBTRACT OFF AND INDICIE BECAUSE THIS IS THE VALUE SET TO NAN AS THE SPIKE WAS INITIATED IN THE PREVIOUS TIME STEP. 
                                                 th_inf_via_Vmeasure,
                                                 dt, 
                                                 a_spike_component_of_threshold, 
                                                 b_spike_component_of_threshold),
                                           x0=[a_voltage_guess,b_voltage_guess])

            a_voltage_comp_of_thr_from_fitab=fit_ab_vcomp_from_noise[0]
            b_voltage_comp_of_thr_from_fitab=fit_ab_vcomp_from_noise[1]

            logging.info("spike components %s %s", str(a_spike_component_of_threshold), str(b_spike_component_of_threshold))
            logging.info("voltage components %s %s", str(a_voltage_comp_of_thr_from_fitab), str(b_voltage_comp_of_thr_from_fitab))

            fit_ab_vcomp_th_from_thr_from_noise = fmin(func=fit_avoltage_bvoltage_th, 
                                                       args=(noise1_data['voltage'], 
                                                             noise_El_list, 
                                                             spike_cut_length_NODELTAV, 
                                                             noise1_ind_wo_test_pulse_removed, #NOTE THAT IF YOU WANT TO USE THIS TO GET A VOLTAGE WITHIN THE FUNCTION YOU NEED TO SUBTRACT OFF AND INDICIE BECAUSE THIS IS THE VALUE SET TO NAN AS THE SPIKE WAS INITIATED IN THE PREVIOUS TIME STEP. 
                                                             dt, 
                                                             a_spike_component_of_threshold, 
                                                             b_spike_component_of_threshold),
                                                       x0=[a_voltage_guess,b_voltage_guess, th_inf_via_Vmeasure])

            a_voltage_comp_of_thr_from_fitabth=fit_ab_vcomp_th_from_thr_from_noise[0]
            b_voltage_comp_of_thr_from_fitabth=fit_ab_vcomp_th_from_thr_from_noise[1]
            th_inf_fit_w_v_comp_of_th=fit_ab_vcomp_th_from_thr_from_noise[2]    
            th_inf_fit_w_v_comp_of_th_from0=th_inf_fit_w_v_comp_of_th-El_noise
            logging.info("spike components %s %s", str(a_spike_component_of_threshold), str(b_spike_component_of_threshold))
            logging.info("voltage components %s %s %s %s", str(a_voltage_comp_of_thr_from_fitabth), str(b_voltage_comp_of_thr_from_fitabth), 'fit threshold', str(th_inf_fit_w_v_comp_of_th))


    else:
        a_spike_component_of_threshold=None
        b_spike_component_of_threshold=None
        a_voltage_comp_of_thr_from_fitab=None
        b_voltage_comp_of_thr_from_fitab=None 
        a_voltage_comp_of_thr_from_fitabth=None
        b_voltage_comp_of_thr_from_fitabth=None
        th_inf_fit_w_v_comp_of_th_from0=None
        th_inf_fit_w_v_comp_of_th=None

    #--------------------------------------------------------------------------     
    #------------------------ MLIN calculations--------------------------------
    #--------------------------------------------------------------------------

    #TODO: probably want to use more than just one square pulse for this distribution
    STLS_voltage=maximum_subthreshold_long_square_data['voltage'][0][maximum_subthreshold_long_square_data['start_idx'][0]:]
    STLS_current=maximum_subthreshold_long_square_data['current'][0][maximum_subthreshold_long_square_data['start_idx'][0]:]
    (var_of_section, sv_for_expsymm, tau_from_AC)=MLIN(STLS_voltage, STLS_current, R_test_list_mean, C_test_list_mean, dt, 
                                                       MAKE_PLOT=MAKE_PLOT, 
                                                       SHOW_PLOT=SHOW_PLOT, 
                                                       BLOCK=False, 
                                                       PUBLICATION_PLOT=False)
    if SAVE_FIG:
        tag='MLIN.png'
        tag_plot(tag)
        plt.savefig(os.path.join(save_figure_path,tag), format='png')
        plt.close() 

    #--------------------------------------------------------------------------     
    #------------------------ make output dictionaries-------------------------
    #--------------------------------------------------------------------------

    #TODO: find out how many are the max number of all_passing_sweeps. 
    El_noise_1=[None, None, None, None, None]
    WFS_noise_1=[None, None, None, None, None]
    RTP_noise_1=[None, None, None, None, None]
    sweep_noise_1=[None, None, None, None, None]
    spike_ind_noise_1=[None, None, None, None, None]

    def fill_in_lists(out_list, data_list):
        '''note since the input is a list shouldnt need to return anything (pass by reference)'''
        for ii in range(len(data_list)):
            out_list[ii]=data_list[ii] 
    fill_in_lists(El_noise_1, noise_El_list)
    fill_in_lists(sweep_noise_1, noise1_sweeps)
    fill_in_lists(spike_ind_noise_1, noise1_ind_wo_test_pulse_removed)

    #--initialize output dictionaries
    for_reference_dict={}
    for_use_dict={}

    for_reference_dict['dt_used_for_preprocessor_calculations']=dt
    #for_reference_dict['optional_methods']=self.optional_methods
    for_reference_dict['sweep_properties']={'noise1': 
                                                {'1':{'El': El_noise_1[0], 'sweep_num':sweep_noise_1[0], 'spike_ind': spike_ind_noise_1[0]},
                                                 '2':{'El': El_noise_1[1], 'sweep_num':sweep_noise_1[1], 'spike_ind': spike_ind_noise_1[1]},
                                                 '3':{'El': El_noise_1[2], 'sweep_num':sweep_noise_1[2], 'spike_ind': spike_ind_noise_1[2]},
                                                 '4':{'El': El_noise_1[3], 'sweep_num':sweep_noise_1[3], 'spike_ind': spike_ind_noise_1[3]},
                                                 '5':{'El': El_noise_1[4], 'sweep_num':sweep_noise_1[4], 'spike_ind': spike_ind_noise_1[4]}},
                                            'ramp': {'sweep_num':ramp_sweeps},
                                            'subthreshold_short_square': {'sweep_num':maximum_subthreshold_short_square_sweeps},
                                            'suprathreshold_short_square': {'R_testpulsesweep_num':minimum_suprathreshold_short_square_sweeps},
                                            'max_subthresh_long_square': {'sweep_num':maximum_subthreshold_long_square_sweeps},
                                            'multi_short_square': {'sweep_num':ssq_triple_sweeps}}

    for_reference_dict['El']={'El_noise': {'measured': {'mean':El_noise, 'list':noise_El_list, 'dependencies':None}},
                              'El_ramp': {'value':El_ramp, 'dependencies': None},
                              'El_subthreshold_blip': {'value':El_subthreshold_blip, 'dependencies':None}, 
                              'El_suprathreshold_blip': {'value':El_suprathreshold_blip, 'dependencies':None},
                              'El_max_subth_long_square': {'value':El_max_subth_long_square, 'dependencies':None}} 

    for_reference_dict['resistance']={#'R_lssq_Wrest':{'mean': R_lssq_wrest_mean, 'list': R_lssq_wrest_list, 'dependencies': 'from subthreshold (no spike cutting) noise'},        
                                      'R_from_lims':{'value':cell_properties['ri']*1e6},
                                      'R_test_list': {'mean':R_test_list_mean, 'list': R_test_list},
                                      'R_fit_ASC_and_R':{'mean':R_from_ASGLM, 'list': best_R_fit_ascR}}

    for_reference_dict['capacitance']={#'C_lssq_Wrest': {'mean':C_lssq_wrest_mean, 'list':C_lssq_wrest_list, 'dependencies': 'from subthreshold (no spike cutting) noise'},
                                       'C_from_lims':{'value': (cell_properties['tau']*1e-3)/(cell_properties['ri']*1e6)},
                                       'C_test_list': {'mean':C_test_list_mean, 'list': C_test_list}}
    
    for_reference_dict['spike_cut_length']={'no deltaV shift':{'length':spike_cut_length_NODELTAV, 
                                                               'slope':slope_at_min_expVar_list_NODELTAV, 
                                                               'intercept':intercept_at_min_expVar_list_NODELTAV, 
                                                               'dependencies':None}}

    for_reference_dict['spike_cutting']={'NOdeltaV': {'cut_length':spike_cut_length_NODELTAV, 'slope':slope_at_min_expVar_list_NODELTAV, 'intercept':intercept_at_min_expVar_list_NODELTAV, 'dependencies': None}}

    for_reference_dict['asc']={'k': best_k_pair_fit_ascR, 'amp':asc_amp_from_ASGLM, 'dependencies': 'Cap and res from least squares'}

    for_reference_dict['th_inf']={'via_Vmeasure':{'value':th_inf_via_Vmeasure, 'from_zero':th_inf_via_Vmeasure_from0, 'dependencies':'measured from suprathreshold blip'},
                                  'fit_with_v_comp_of_th':{'value':th_inf_fit_w_v_comp_of_th, 'from_zero':th_inf_fit_w_v_comp_of_th_from0}}

    for_reference_dict['threshold_adaptation']={'a_spike_component_of_threshold':a_spike_component_of_threshold,
                                                'b_spike_component_of_threshold':b_spike_component_of_threshold,
                                                'a_voltage_comp_of_thr_from_fitab':a_voltage_comp_of_thr_from_fitab,
                                                'b_voltage_comp_of_thr_from_fitab':b_voltage_comp_of_thr_from_fitab,
                                                'a_voltage_comp_of_thr_from_fitabth':a_voltage_comp_of_thr_from_fitabth,
                                                'b_voltage_comp_of_thr_from_fitabth':b_voltage_comp_of_thr_from_fitabth}
    for_reference_dict['MLIN']={'var_of_section':var_of_section,
                                'sv_for_expsymm':sv_for_expsymm,
                                'tau_from_AC':tau_from_AC}
    logging.info("finished")

    return for_reference_dict

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("nwb_file")
    parser.add_argument("sweep_list_file")
    parser.add_argument("output_json")
    parser.add_argument("--dt", default=DEFAULT_DT)
    parser.add_argument("--bessel", default=DEFAULT_BESSEL)
    parser.add_argument("--cut", default=DEFAULT_CUT)

    args = parser.parse_args()

    sweep_list = ju.read(args.sweep_list_file)
    
    values = preprocess_neuron(args.nwb_file, sweep_list)

    ju.write(args.output_json, values)
    

if __name__ == "__main__": main()



from __future__ import division
import multiprocessing as mp
import logging
import functools
import os

from six import iteritems
import SimpleITK as sitk
import numpy as np

from base_subimage import run_subimage
from image_series_gridder import ImageSeriesGridder
from downsampling_utilities import downsample_average
from image_utilities import image_from_array, write_volume


#==============================================================================

    
def run_cav(grid_prefix, accumulator_prefix, **input_data):

    paths = []

    isg = ImageSeriesGridder(**input_data)
    isg.setup_subimages()

    isg.build_coarse_grids()

    cb = functools.partial(write_volume, name='cav_tracer_10', prefix=accumulator_prefix, paths=paths)
    isg.accumulator_to_numpy('cav_tracer', cb)

    cb = functools.partial(write_volume, name='sum_pixels_10', prefix=accumulator_prefix, paths=paths)
    isg.accumulator_to_numpy('sum_pixels', cb)

    isg.make_ratio_volume('cav_tracer', 'sum_pixels', 'cav_density')
    isg.volumes['cav_density'] = image_from_array(isg.volumes['cav_density'], isg.out_spacing)
    write_volume(isg.volumes['cav_density'], name='cav_density_10', prefix=grid_prefix, paths=paths)
    del isg.volumes['cav_density']

    isg.volumes['data_mask'] = isg.volumes['sum_pixels'] / np.amax(isg.volumes['sum_pixels'])
    del isg.volumes['sum_pixels']
    isg.volumes['data_mask'][isg.volumes['data_mask'] > 0.5] = 1
    isg.volumes['data_mask'][isg.volumes['data_mask'] < 0.5] = 0
    isg.volumes['data_mask'] = image_from_array(isg.volumes['data_mask'], isg.out_spacing)
    write_volume(isg.volumes['data_mask'], name='data_mask_10', prefix=accumulator_prefix, paths=paths)
    del isg.volumes

    return paths


#==============================================================================








































import sys
import argparse
import logging
import os

from xml.etree.ElementTree import Element, SubElement, Comment, tostring
from xml.dom import minidom

import SimpleITK as sitk
import numpy as np
from six import iteritems

from allensdk.internal.core.lims_pipeline_module import PipelineModule, run_module
from allensdk.internal.mouse_connectivity.tissuecyte_stitching.stitcher import Stitcher
from allensdk.internal.mouse_connectivity.tissuecyte_stitching.tile import Tile
import allensdk.core.json_utilities as ju

# TODO this ought to be installed with the actual python build? 
# need to consult with sysadmins/refactor jp2 project build
sys.path.append('/shared/bioapps/itk/itk_shared/jp2/build')
import jpeg_twok


logging.getLogger().setLevel(logging.INFO)
logging.captureWarnings(True)


def get_missing_tile_paths(missing_tiles):

    paths = []

    for index, path in iteritems(missing_tiles):
        spath = ','.join(map(str, path))
        logging.info('writing missing tile path for tile {0} as {1}'.format(index, spath))
        paths.append(spath)

    return paths


def read_image(file_name):
    logging.info('reading image from {0}'.format(file_name))
    image = sitk.ReadImage(str(file_name))
    return np.flipud(sitk.GetArrayFromImage(image)).T


def normalize_image_by_median(image):

    median = np.median(image)

    if median != 0:
        image = np.divide(median, image)
        image[np.isnan(image)] = 0
        image[np.isinf(image)] = 0

    return image


def load_average_tile(path):
    tile = read_image(path)
    return normalize_image_by_median(tile)


def get_average_tiles(average_tile_paths):

    average_tiles = {}    
    for key, path in iteritems(average_tile_paths):
        key = int(key) - 1

        try:
            average_tiles[key] = load_average_tile(path)
            logging.info('found average tile for channel {0} (zero-indexed)'.format(key))        
        except(IOError, OSError, RuntimeError) as err:
            average_tiles[key] = None
            logging.info('did not find average tile for channel {0} (zero-indexed)'.format(key))
        
    return average_tiles


def generate_tiles(tiles):

    for tile_params in tiles:
        tile = tile_params.copy()

        try:
            tile['image'] = read_image(tile['path'])
            tile['is_missing'] = False
        except (IOError, OSError, RuntimeError) as err:
            tile['image'] = None
            tile['is_missing'] = True
        
        tile['channel'] = tile['channel'] - 1

        tile_obj = Tile(**tile)
        del tile
        yield tile_obj


def write_output(arr, spacing, path):
    jpeg_twok.write(arr, path)


def main():

    output_json = args.output_json
    output_directory = os.path.dirname(output_json)

    slice_path = os.path.join(output_directory, data['slice_fname'])
    
    tiles = generate_tiles(data['tiles'])
    average_tiles = get_average_tiles(data['average_tile_paths'])

    stitcher = Stitcher(data['image_dimensions'], tiles, average_tiles, data['channels'])
    image, missing = stitcher.run()
    del tiles
    missing_tile_paths = get_missing_tile_paths(missing)

    write_output(np.ascontiguousarray(image), data['spacing'], slice_path)

    module_outputs = {'slice_fname': slice_path, 
                      'missing_tile_paths': missing_tile_paths}
    ju.write(output_json, module_outputs)
        

if __name__ == '__main__':

    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    parser = argparse.ArgumentParser()
    parser.add_argument('input_json', type=str)
    parser.add_argument('output_json', type=str)
    args = parser.parse_args()

    data = ju.read(args.input_json)

    main()



from allensdk.core.nwb_data_set import NwbDataSet
from scipy import signal
import numpy as np

def load_sweep(file_name, sweep_number, desired_dt=None, cut=0, bessel=False):
    '''load a data sweep and do specified data processing.
    Inputs:
        file_name: string
            name of .nwb data file
        sweep_number: 
            number specifying the sweep to be loaded
        desired_dt: 
            the size of the time step the data should be subsampled to
        cut:
            indicie of which to start reporting data (i.e. cut off data before this indicie)
        bessel: dictionary
            contains parameters 'N' and 'Wn' to implement standard python bessel filtering
    Returns:
        dictionary containing
            voltage: array
            current: array
            dt: time step of the returned data
            start_idx: the index at which the first stimulus starts (excluding the test pulse)
    '''
    ds = NwbDataSet(file_name)
    data = ds.get_sweep(sweep_number)

    data["dt"] = 1.0 / data["sampling_rate"]

    if cut > 0:
        data["response"] = data["response"][cut:]
        data["stimulus"] = data["stimulus"][cut:]        

    if bessel:
        sample_freq = 1. / data["dt"]
        filt_coeff = (bessel["freq"]) / (sample_freq / 2.) # filter fraction of Nyquist frequency
        b, a = signal.bessel(bessel["N"], filt_coeff, "low")
        data['response'] = signal.filtfilt(b, a, data['response'], axis=0)

    if desired_dt is not None:
        if data["dt"] != desired_dt:
            data["response"] = subsample_data(data["response"], "mean", data["dt"], desired_dt)
            data["stimulus"] = subsample_data(data["stimulus"], "mean", data["dt"], desired_dt)
            data["start_idx"] = int(data["index_range"][0] / (desired_dt / data["dt"]))
            data["dt"] = desired_dt

    if "start_idx" not in data:
        data["start_idx"] = data["index_range"][0]

    return {
        "voltage": data["response"],
        "current": data["stimulus"],
        "dt": data["dt"],
        "start_idx": data["start_idx"]
        }


def load_sweeps(file_name, sweep_numbers, dt=None, cut=0, bessel=False):
    '''load sweeps and do specified data processing.
    Inputs:
        file_name: string
            name of .nwb data file
        sweep_numbers: 
            sweep numbers to be loaded
        desired_dt: 
            the size of the time step the data should be subsampled to
        cut:
            indicie of which to start reporting data (i.e. cut off data before this indicie)
        bessel: dictionary
            contains parameters 'N' and 'Wn' to implement standard python bessel filtering
    Returns:
        dictionary containing
            voltage: list of voltage trace arrays
            current: list of current trace arrays
            dt: list of time step corresponding to each array of the returned data
            start_idx: list of the indicies at which the first stimulus starts (excluding 
                the test pulse) in each returned sweep
    '''
    data = [ load_sweep(file_name, sweep_number, dt, cut, bessel) for sweep_number in sweep_numbers ]

    return {
        'voltage': [ d['voltage'] for d in data ],
        'current': [ d['current'] for d in data ],
        'dt': [ d['dt'] for d in data ],
        'start_idx': [ d['start_idx'] for d in data ],
        } 


def subsample_data(data, method, present_time_step, desired_time_step):
    if present_time_step > desired_time_step:
        raise Exception("you desired time step is smaller than your present time step")

    # number of elements to average over
    n = int(desired_time_step / present_time_step)

    data_subsampled = None

    if method == "mean":
        # if n does not divide evenly into the length of the array, crop off the end
        end = n * int(len(data) / n)
            
        return np.mean(data[:end].reshape(-1,n), 1)

    raise Exception("unknown subsample method: %s" % (method))


import numpy as np
import numpy.fft as npft

# TODO: license
# TODO: normalize function call names
# TODO: document functions

def create_basis_IPSP(neye,ncos,kpeaks,ks,DTsim,t0,I_stim,nkt,flag_exp,npcut):
    
    kbasprs = {}
    kbasprs['neye'] = neye #No of 'identity' basis vectors near time of spike
    kbasprs['ncos'] = ncos #No of raised-cosines to use
    kbasprs['kpeaks'] = kpeaks  #Position of first and last bump
    kbasprs['b'] = 0.1 #Offset for non-linear scaling
    kbasprs['ks'] = ks
        
    gg0 = makeFitStruct_GLM(DTsim,kbasprs,nkt,flag_exp)
    
    #Create spike-stim with which to convolve post spike filter
    spike_stim = np.zeros(np.shape(I_stim))
    for kk in range(len(t0)):
        spind = int(t0[kk])
        #print int(t0[kk]), spind-190000
        spike_stim[spind]=1.0
    
    ##Convolve temporal basis functions with spike-stim
    c = np.zeros((len(spike_stim),ncos))
    for jj in range(ncos):
        basisfilt = gg0['ktbas'][:,jj]
        bconv = np.convolve(spike_stim,np.flipud(basisfilt),'full')
        c[:,jj] = bconv[range(len(spike_stim))]        
    
    basis_IPSP = c;
    
    return basis_IPSP, gg0

def makeFitStruct_GLM(dtsim,kbasprs,nkt,flag_exp):
    
    gg = {}
    gg['k'] = []
    gg['dc'] = 0
    gg['kt'] = np.zeros((nkt,1))
    gg['ktbas'] = []
    gg['kbasprs'] = kbasprs
    gg['dt'] = dtsim
    
    nkt = nkt
    if flag_exp==0:    
        ktbas = makeBasis_StimKernel(kbasprs,nkt)
    else:
        ktbas = makeBasis_StimKernel_exp(kbasprs,nkt)
    
    gg['ktbas'] = ktbas
    gg['k'] = gg['ktbas']*gg['kt']
    
    return gg

def makeBasis_StimKernel(kbasprs,nkt):
    
    neye = kbasprs['neye']
    ncos = kbasprs['ncos']
    kpeaks = kbasprs['kpeaks']
    kdt = 1
    b = kbasprs['b']
    
    yrnge = nlin(kpeaks + b*np.ones(np.shape(kpeaks)))
    #db = np.diff(yrnge)/(ncos-1)
    db = (yrnge[-1]-yrnge[0])/(ncos-1)
    ctrs = yrnge
    mxt = invnl(yrnge[ncos-1]+2*db)-b
    print mxt
    kt0 = np.arange(0,mxt,kdt)    
    nt = len(kt0)
    e1 = np.tile(nlin(kt0+b*np.ones(np.shape(kt0))),(ncos,1))
    e2 = np.transpose(e1)    
    e3 = np.tile(ctrs,(nt,1))
    kbasis0 = []
    for kk in range(ncos):    
        kbasis0.append(ff(e2[:,kk],e3[:,kk],db))
         
    
    #Concatenate identity vectors
    nkt0 = np.size(kt0,0)
    a1 = np.concatenate((np.eye(neye), np.zeros((nkt0,neye))),axis=0)
    a2 = np.concatenate((np.zeros((neye,ncos)),np.array(kbasis0).T),axis=0)
    kbasis = np.concatenate((a1,a2),axis=1)
    kbasis = np.flipud(kbasis)
    nkt0 = np.size(kbasis,0)
    
    if nkt0 < nkt:
        kbasis = np.concatenate((np.zeros((nkt-nkt0,ncos+neye)),kbasis),axis=0)
    elif nkt0 > nkt:
        kbasis = kbasis[-1-nkt:-1,:] 

    kbasis = normalizecols(kbasis)

    return kbasis    


def makeBasis_StimKernel_exp(kbasprs,nkt):
    ks = kbasprs['ks']
    b = kbasprs['b']
    x0 = np.arange(0,nkt)
    kbasis = np.zeros((nkt,len(ks)))
    for ii in range(len(ks)):
        kbasis[:,ii] = invnl(-ks[ii]*x0) #(1.0/ks[ii])*
    
    kbasis = np.flipud(kbasis)     
    return kbasis  
    
def nlin(x):
    eps = 1e-20
    return np.log(x+eps)
    
def invnl(x):
    eps = 1e-20
    return np.exp(x)-eps
    
def ff(x,c,dc):
    rowsize = np.size(x,0)
    m = []
    for i in range(rowsize): 
        xi = x[i]
        ci = c[i]
        val=(np.cos(np.max([-pi,np.min([pi,(xi-ci)*pi/dc/2])]))+1)/2    
        m.append(val)
        
    return np.array(m)
    
def normalizecols(A):
    
    B = A/np.tile(np.sqrt(sum(A**2,0)),(np.size(A,0),1))
    
    return B
    
def sameconv(A,B):
    
    am = np.size(A)
    bm = np.size(B)
    nn = am+bm-1
    
    q = npft.fft(A,nn)*npft.fft(np.flipud(B),nn)
    p = q
    G = npft.ifft(p)
    G = G[range(am)]
    
    return G




import json, sys, os
import logging
import argparse
import allensdk.core.json_utilities as ju


SHORT_SQUARE = 'Short Square'
SHORT_SQUARE_60 = 'Short Square - Hold -60mv'
SHORT_SQUARE_80 = 'Short Square - Hold -80mv'
LONG_SQUARE = 'Long Square'
RAMP = 'Ramp'
NOISE1 = 'Noise 1'
NOISE2 = 'Noise 2'
SHORT_SQUARE_TRIPLE = 'Short Square - Triple'
RAMP_TO_RHEO = 'Ramp to Rheobase'


class MissingSweepException( Exception ): pass

def get_sweep_numbers(sweep_list):
    return [ s['sweep_number'] for s in sweep_list]


def get_sweeps_by_name(sweeps, sweep_type):
    if isinstance(sweeps, dict):
        return [ s for sn,s in sweeps.iteritems() if s[u'ephys_stimulus'][u'ephys_stimulus_type'][u'name'] == sweep_type ]    
    else:
        return [ s for s in sweeps if s[u'ephys_stimulus'][u'ephys_stimulus_type'][u'name'] == sweep_type ]


def find_ranked_sweep(sweep_list, key, reverse=False):
    if sweep_list:
        sorted_sweep_list = sorted(sweep_list, key=lambda x: x[key], reverse=reverse)
        
        out_sweeps = [ sorted_sweep_list[0] ]
 
        for i in xrange(1,len(sweep_list)):
            if sorted_sweep_list[i][key] == out_sweeps[0][key]:
                out_sweeps.append(sorted_sweep_list[i])
            else:
                break
        
        return get_sweep_numbers(out_sweeps)
    else:
        return []


def organize_sweeps_by_name(sweeps, name):
    sweep_list = sorted(get_sweeps_by_name(sweeps, name), key=lambda x: x['sweep_number'])

    subthreshold_list = [ s for s in sweep_list if s.get('num_spikes',None) in [0, None] ]    
    suprathreshold_list = [ s for s in sweep_list if s.get('num_spikes',None) > 0 ]    

    return {
        'all': get_sweep_numbers(sweep_list),
        'subthreshold': get_sweep_numbers(subthreshold_list),
        'suprathreshold': get_sweep_numbers(suprathreshold_list),
        'maximum_subthreshold': find_ranked_sweep(subthreshold_list, 'stimulus_amplitude', reverse=True),
        'minimum_suprathreshold': find_ranked_sweep(suprathreshold_list, 'stimulus_amplitude')
        #'maximum_subthreshold': find_ranked_sweep(subthreshold_list, 'stimulus_absolute_amplitude', reverse=True),
        #'minimum_suprathreshold': find_ranked_sweep(suprathreshold_list, 'stimulus_absolute_amplitude')
        }

def find_long_square_sweeps(sweeps):
    out = organize_sweeps_by_name(sweeps, LONG_SQUARE)
    return out


def find_ramp_to_rheo_sweeps(sweeps):
    out = organize_sweeps_by_name(sweeps, RAMP_TO_RHEO)
    return out


def find_short_square_sweeps(sweeps):
    '''
    Find 1) all of the subthreshold short square sweeps
         2) all of the superthreshold short square sweeps
         3) the subthresholds short square sweep with maximum stimulus amplitude
    '''

    out = organize_sweeps_by_name(sweeps, SHORT_SQUARE)
    out60 = organize_sweeps_by_name(sweeps, SHORT_SQUARE_60)
    out80 = organize_sweeps_by_name(sweeps, SHORT_SQUARE_80)
    out_triple = organize_sweeps_by_name(sweeps, SHORT_SQUARE_TRIPLE)

    out['all_60'] = out60['all']
    out['all_80'] = out80['all']
    out['triple'] = out_triple['all']
    
    if len(out['maximum_subthreshold']) == 0: 
        raise MissingSweepException("No maximum subthreshold short square")

    if len(out['minimum_suprathreshold']) == 0:
        raise MissingSweepException("No minimum suprathreshold short square")

    return out


def find_ramp_sweeps(sweeps):
    '''
    Find 1) all ramp sweeps
         2) all subthreshold ramps
         3) all superthreshold ramps
    '''
    out = organize_sweeps_by_name(sweeps, RAMP)

    return out
    

def find_noise_sweeps(sweeps):
    '''
    Find 1) the noise1 sweeps
         2) the noise2 sweeps
         4) all noise sweeps
    '''

    noise1 = organize_sweeps_by_name(sweeps, NOISE1)
    noise2 = organize_sweeps_by_name(sweeps, NOISE2)
    
    all_noise_sweeps = sorted(noise1['all'] + noise2['all'])

    out = {
        'all': all_noise_sweeps,
        'noise1': noise1['all'],
        'noise2': noise2['all']
    }

    num_noise1_sweeps = len(out['noise1'])
    num_noise2_sweeps = len(out['noise2'])
    
    required_noise1_sweeps = 2
    required_noise2_sweeps = 2

    if num_noise1_sweeps < required_noise1_sweeps:
        raise MissingSweepException("not enough noise1 sweeps (%d/%d)" % (num_noise1_sweeps, required_noise1_sweeps))
 
    if num_noise2_sweeps < required_noise2_sweeps:
        raise MissingSweepException("not enough noise2 sweeps (%d/%d)" % (num_noise2_sweeps, required_noise2_sweeps))
         
    return out


def find_sweeps(sweep_list):

    sweep_index = { s['sweep_number']: s for s in sweep_list }

    data = {}
    ssq_data = find_short_square_sweeps(sweep_index)
    data.update(ssq_data)

    lsq_data = find_long_square_sweeps(sweep_index)
    data.update(lsq_data)

    ramp_data = find_ramp_sweeps(sweep_index)
    data.update(ramp_data)

    r2r_data = find_ramp_to_rheo_sweeps(sweep_index)
    data.update(r2r_data)

    noise_data = find_noise_sweeps(sweep_index)
    data.update(noise_data)

    return data, sweep_index
    

def parse_arguments():
    parser = argparse.ArgumentParser(description='find relevant sweeps from a sweep catalog')

    parser.add_argument('sweep_list_file', help='json file containing a list of sweeps for a cell')
    parser.add_argument('output_file', help='output json data config file')

    args = parser.parse_args()

    try:
        if not os.path.exists(args.sweep_list_file):
            raise Exception("sweep list file (%s) does not exist" % args.sweep_file)

    except Exception, e:
        parser.print_help()
        sys.exit(1)

    return args


def main():
    args = parse_arguments()

    sweep_list = ju.read(args.sweep_list_file)

    data = find_sweeps(sweep_list)

    ju.write(args.output_file, data)

    if len(errs > 0):
        for err in errs:
            logging.error(err)
        sys.exit(1)

if __name__ == "__main__":  main()



import allensdk.internal.model.biophysical.ephys_utils as ephys_utils
import logging
import numpy as np
import pandas as pd

_passive_fit_log = logging.getLogger(
    'allensdk.model.biophysical.passive_fitting.preprocess')

def get_passive_fit_data(cap_check_sweeps, data_set):
    bridge_balances = [s['bridge_balance_mohm'] for s in cap_check_sweeps]
    bridge_avg = np.array(bridge_balances).mean()
    _passive_fit_log.debug("bridge avg {:.2f}".format(bridge_avg))

    initialized = False
    for idx, s in enumerate(cap_check_sweeps):
        v, i, t = ephys_utils.get_sweep_v_i_t_from_set(data_set,
                                                       s['sweep_number'])
        if v is None:
            continue
        up_idxs, down_idxs = get_cap_check_indices(i)

        down_idx_interval = down_idxs[1] - down_idxs[0]
        skip_count = 0
        for j in range(len(up_idxs)):
            if j == 0:
                avg_up = v[(up_idxs[j] - 400):down_idxs[j + 1]]
                avg_down = v[(down_idxs[j] - 400):up_idxs[j]]
            elif j == len(up_idxs) - 1:
                avg_up = avg_up + v[(up_idxs[j] - 400):-2]
                avg_down = avg_down + v[(down_idxs[j] - 400):up_idxs[j]]
            else:
                avg_up = avg_up + v[(up_idxs[j] - 400):down_idxs[j + 1]]
                avg_down = avg_down + v[(down_idxs[j] - 400):up_idxs[j]]
        avg_up /= len(up_idxs) - skip_count
        avg_down /= len(up_idxs) - skip_count
        if not initialized:
            grand_up = avg_up - avg_up[0:400].mean()
            grand_down = avg_down - avg_down[0:400].mean()
            initialized = True
        else:
            grand_up = grand_up + (avg_up - avg_up[0:400].mean())
            grand_down = grand_down + (avg_down - avg_down[0:400].mean())
    grand_up /= len(cap_check_sweeps)
    grand_down /= len(cap_check_sweeps)

    t = 0.005 * np.arange(len(grand_up)) # in ms, assumes 200kHz sampling rate]

    grand_up_data = np.column_stack((t, grand_up))
    grand_down_data = np.column_stack((t, grand_down))

    grand_diff = (grand_up + grand_down) / grand_up
    avg_grand_diff = pd.rolling_mean(pd.Series(grand_diff, index=t), 100)
    threshold = 0.2
    start_index = np.flatnonzero(t >= 4.0)[0]
    escape_indexes = np.flatnonzero(np.abs(avg_grand_diff.values[start_index:]) > threshold) + start_index
    if len(escape_indexes) < 1:
        escape_index = len(t) - 1
    else:
        escape_index = escape_indexes[0]
    escape_t = t[escape_index]

    return {
        'grand_up': grand_up_data,
        'grand_down': grand_down_data,
        'escape_t': escape_t,
        'bridge_avg': bridge_avg
        }

def get_cap_check_indices(i):
    # Assumes that there is a test pulse followed by the stimulus pulses (downward first)
    di = np.diff(i)
    up_idx = np.flatnonzero(di > 0)
    down_idx = np.flatnonzero(di < 0)

    return up_idx[2::2], down_idx[1::2]


def main(): 
    pass
if __name__ == "__main__": main()



import functools
import logging

import numpy as np
import SimpleITK as sitk
from scipy.ndimage.interpolation import zoom

from image_sheet import ImageSheet
from projection_functions import max_projection
from volume_projector import VolumeProjector
import visualization_utilities as vis


def max_cb(max_sheet, depth_sheet, volume, axis, *a, **k):
    m, d = max_projection(volume, axis)
    max_sheet.append(m.T)
    depth_sheet.append(d.T)


def apply_colormap(image, colormap):
    color_image = colormap(image)
    color_image[:, :, -1] = image
    return color_image


def blend_with_background(image, background):

    for ii in xrange(3):
        image[:, :, ii] = np.multiply(np.squeeze(image[:, :, ii]), 
                                      np.squeeze(image[:, :, -1]))
        image[:, :, ii] += np.multiply(np.squeeze(background), 
                                       np.squeeze(1.0 - image[:, :, -1]))

    image[:, :, -1] = 1
    return image


def do_blur(image, blur):

    for ii in xrange(3):
        im = sitk.GetImageFromArray(image[:, :, ii])
        im = sitk.DiscreteGaussian(im, blur)
        image[:, :, ii] = sitk.GetArrayFromImage(im)
    
    return image


def handle_output_image(sheet, out_image, colormap, nsteps):

    sheet = sheet.copy()
    whole_sheet = sheet.get_output(-1)

    if out_image['blur'] > 0.0:
        logging.info('applying a gaussian blur with variance: {0:2.2f}'.format(out_image['blur']))
        whole_sheet = sitk.GetImageFromArray(whole_sheet)
        whole_sheet = sitk.DiscreteGaussian(whole_sheet, out_image['blur'])
        whole_sheet = sitk.GetArrayFromImage(whole_sheet)

    if out_image['scale'] != 1:
        whole_sheet = zoom(whole_sheet, zoom=out_image['scale'], order=1)

    whole_sheet = apply_colormap(whole_sheet, colormap)

    if out_image['background'] is not None:
        whole_sheet = blend_with_background(whole_sheet, out_image['background'])
    else:
        whole_sheet = blend_with_background(whole_sheet, np.zeros_like(whole_sheet)[:, :, -1])

    whole_sheet = np.around(whole_sheet * 255).astype(np.uint8)
    out_image['write'](whole_sheet[:, :, :-1])


def simple_rotation(from_axis, to_axis, start, end, nsteps):

    angles = np.linspace(start * np.pi, end * np.pi, nsteps, endpoint=False)
    from_axes = [from_axis] * nsteps
    to_axes = [to_axis] * nsteps

    return from_axes, to_axes, angles


def run(volume, imin, imax, rotations, colormap):

    volume = vis.sitk_safe_ln(volume)

    ln_imin = np.log(imin) if imin != 0 else -np.inf
    ln_imax = np.log(imax) if imax != 0 else np.inf

    volume = sitk.IntensityWindowing(volume, ln_imin, ln_imax, 0.0, 1.0)

    for rotation in rotations:
        
        max_sheet = ImageSheet()
        depth_sheet = ImageSheet()

        vp = VolumeProjector.fixed_factory(volume, rotation['window_size'])
        callback = functools.partial(max_cb, max_sheet, depth_sheet, **rotation['projection_parameters'])

        rot = rotation['rotation_parameters']
        from_axes, to_axes, angles = simple_rotation(**rot)

        for response in vp.rotate_and_extract(from_axes, to_axes, angles, callback):
            pass

        rotation['write_depth_sheet'](depth_sheet.get_output(-1))

        for out_image in rotation['output_images']:
            handle_output_image(max_sheet, out_image, colormap, rot['nsteps'])



########################################################################
# library code
import math
import argparse
import sys
import numpy as np
from scipy.spatial.distance import euclidean
import neuron_morphology.swc as swc
import skimage.draw

#import allensdk.core.json_utilities as json


def calculate_centroid(x, y):
    ''' Calculates the center of a polygon, using weighted averages
    of vertex locations
    '''
    assert len(x) == len(y), "Vertex arrays are of incorrect shape"
    tot_len = 0.0
    tot_x = 0.0
    tot_y = 0.0
    for i in range(len(x)):
        x0 = x[i-1]
        y0 = y[i-1]
        x1 = x[i]
        y1 = y[i]
        seg_len = euclidean((x0, y0), (x1, y1))
        tot_len += seg_len
        tot_x += seg_len * x0
        tot_x += seg_len * x1
        tot_y += seg_len * y0
        tot_y += seg_len * y1
    tot_x /= 2.0 * tot_len
    tot_y /= 2.0 * tot_len
    return tot_x, tot_y


def construct_affine(theta):
    tr_rot = [np.cos(theta), np.sin(theta), 0,
              -np.sin(theta), np.cos(theta), 0,
              0, 0, 1,
              0, 0, 0
             ]
    return tr_rot


#def get_pia_wm_rotation_transform(soma_coords, wm_coords, pia_coords, resolution):
#    # get soma position using weighted average of vertices
#    sx, sy = convert_coords_str(soma_coords)
#    soma_x, soma_y = calculate_centroid(sx, sy)
#
#    #pia_proj = project_to_polyline(pia_coords, avg_soma_position)
#    #wm_proj = project_to_polyline(wm_coords, avg_soma_position)
#    px, py, wx, wy = calculate_shortest(soma_x, soma_y, pia_coords, wm_coords)
#    theta = vector_angle((0, 1), np.asarray([px,py]) - np.asarray([wx,wy]))
#    print theta
#
#    depth = euclidean((soma_x, soma_y), (px, py))
#    height = euclidean((wx, wy), (px, py))
#    return theta, resolution*depth, resolution*height, [px, py, wx, wy]


def convert_coords_str(coord_str):
    vals = coord_str.split(',')
    x = np.array(vals[0::2], dtype=float)
    y = np.array(vals[1::2], dtype=float)
    return x, y

def calculate_shortest(soma_x, soma_y, pia, wm):
    """ Calculates shortest distance through a point on the polygon wm
    through the soma coordinates (soma_x, soma_y) and
    through a point on the polygon pia.

    Returns the x,y points in pia and wm that define the endpoints of this
    shortest line.
    """
    pia_xs, pia_ys = convert_coords_str(pia)
    wm_xs, wm_ys = convert_coords_str(wm)
    #################
    # calculate canvas size and make canvas
    width = max(pia_xs)
    width = int(max(width, max(wm_xs)))
    height = max(pia_ys)
    height = int(max(height, max(wm_ys)))

    canvas = np.zeros((height+10, width+10, 3), dtype=np.uint8)
    
    for i in range(1,len(pia_xs)):
        lr, lc = skimage.draw.line(int(pia_ys[i-1]), int(pia_xs[i-1]), int(pia_ys[i]), int(pia_xs[i]))
        canvas[lr,lc,2] = 255

    for i in range(1,len(wm_xs)):
        lr, lc = skimage.draw.line(int(wm_ys[i-1]), int(wm_xs[i-1]), int(wm_ys[i]), int(wm_xs[i]))
        canvas[lr,lc,0] = 255

    # get points in white matter trace
    wp_y, wp_x = np.nonzero(canvas[:,:,0])

    ##################
    # draw an extended line from each wm pix through the soma
    # (there are usually less WM pix than pia pix, so this should be 
    #   faster than iterating through pia pix)
    # make array of blue (pia) channel only to
    pia = canvas[:,:,2]
    # draw line from each wm pix through the soma and into infinity
    #   (line terminates if/when it intersects with pia trace)
    min_dist = None
    min_coord = None    # stores [ pia_x, pia_y, wm_x, wm_y ]
    for i in range(len(wp_x)):
        x0 = wp_x[i]
        y0 = wp_y[i]
        x1 = soma_x
        y1 = soma_y
        #############################################
        # adapted from Bresenham's line algorithm, from rosetacode
        dx = abs(x1 - x0)
        dy = abs(y1 - y0)
        x, y = x0, y0
        sx = -1 if x0 > x1 else 1
        sy = -1 if y0 > y1 else 1
        if dx > dy:
            err = dx / 2.0
            while x >= 0 and x < width and y >= 0 and y < height:
                if pia[y,x] > 0:
                    dist = euclidean((x0, y0), (x, y))
                    if min_dist is None or min_dist > dist:
                        min_dist = dist
                        min_coord = [x, y, x0, y0]
                    break
                err -= dy
                if err < 0:
                    y += sy
                    err += dx
                x += sx
        else:
            err = dy / 2.0
            while x >= 0 and x < width and y >= 0 and y < height:
                if pia[y,x] > 0:
                    dist = euclidean((x0, y0), (x, y))
                    if min_dist is None or min_dist > dist:
                        min_dist = dist
                        min_coord = [x, y, x0, y0]
                    break
                err -= dx
                if err < 0:
                    x += sx
                    err += dy
                y += sy

    if min_dist is None:
        print("Unable to connect pia to WM through soma")
        px = None
        py = None
        wx = None
        wy = None
    else:
        px = min_coord[0]
        py = min_coord[1]
        wx = min_coord[2]
        wy = min_coord[3]
    return px, py, wx, wy


def dist_proj_point_lineseg(p, q1, q2):
    # based on c code from http://stackoverflow.com/questions/849211/shortest-distance-between-a-point-and-a-line-segment
    l2 = euclidean(q1, q2) ** 2
    if l2 == 0:
        return euclidean(p, q1) # q1 == q2 case
    t = max(0, min(1, np.dot(p - q1, q2 - q1) / l2))
    proj = q1 + t * (q2 - q1)
    return euclidean(p, proj), proj


def project_to_polyline(boundary, soma):
    x, y = convert_coords_str(boundary)
    points = zip(x, y)
    dists_projs = [dist_proj_point_lineseg(soma, np.array(q1), np.array(q2))
                 for q1, q2 in zip(points[:-1], points[1:])]
    min_idx = np.argmin(np.array([d[0] for d in dists_projs]))
    return dists_projs[min_idx][1]


def vector_angle(v1, v2):
    return np.arctan2(v2[1], v2[0]) - np.arctan2(v1[1], v1[0])


########################################################################
# pipeline code
import allensdk.internal.core.swc as swc
from allensdk.internal.core.lims_pipeline_module import PipelineModule


def main(jin):
    # per IT-14567, blockface analysis is no longer required
    #########################################################################
    ## analyze blockface image
    #try:
    #    soma = jin["blockface"]["Soma"]["path"]
    #    pia = jin["blockface"]["Pia"]["path"]
    #    wm = jin["blockface"]["White Matter"]["path"]
    #    res = float(jin["blockface"]["Pia"]["resolution"])
    #except:
    #    print("** Error -- missing requisite blockface field(s) in input json")
    #    raise
    #
    ## get soma position using weighted average of vertices
    #try:
    #    sx, sy = convert_coords_str(soma)
    #    soma_x, soma_y = calculate_centroid(sx, sy)
    #except:
    #    print("** Error -- unable to calculate soma information (blockface)")
    #    raise
    #
    ## calculate shortest path
    #try:
    #    px, py, wx, wy = calculate_shortest(soma_x, soma_y, pia, wm)
    ## calculate theta and affine
    #    theta = vector_angle((0, 1), np.asarray([px,py]) - np.asarray([wx,wy]))
    ## calculate soma depth and cortical thickness
    #    depth = res * euclidean((soma_x, soma_y), (px, py))
    #    blk_thickness = res * euclidean((wx, wy), (px, py))
    #except:
    #    print("** Error calculating shortest path (blockface)")
    #    raise
    #
    #blockface = {}
    #blockface["pia_intersect"] = [ px, py ]
    #blockface["wm_intersect"] = [ wx, wy ]
    #blockface["soma_center"] = [ soma_x, soma_y ]
    #blockface["soma_depth_um"] = depth
    #try:
    #    blockface["soma_depth_relative"] = depth / blk_thickness
    #except:
    #    blockface["soma_depth_relative"] = -1.0    # NaN is not friendly to ruby
    #blockface["cort_thickness_um"] = blk_thickness
    #blockface["theta"] = theta
    
    ########################################################################
    # analyze primary (20x) image
    try:
        soma = jin["primary"]["Soma"]["path"]
        pia = jin["primary"]["Pia"]["path"]
        wm = jin["primary"]["White Matter"]["path"]
        res = float(jin["primary"]["Pia"]["resolution"])
    except:
        print("** Error -- missing requisite primary (20x) field(s) in input json")
        raise
    
    # get soma position using weighted average of vertices
    try:
        sx, sy = convert_coords_str(soma)
        soma_x, soma_y = calculate_centroid(sx, sy)
    except:
        print("** Error -- unable to calculate soma information (primary)")
        raise
    try:
        # calculate shortest path
        px, py, wx, wy = calculate_shortest(soma_x, soma_y, pia, wm)
        # calculate theta and affine
        theta = vector_angle((0, 1), np.asarray([px,py]) - np.asarray([wx,wy]))
        tr_rot = construct_affine(theta)
        inv_tr_rot = construct_affine(-theta)
        # calculate soma depth and cortical thickness
        depth = res * euclidean((soma_x, soma_y), (px, py))
        raw_thickness = res * euclidean((wx, wy), (px, py))
    except:
        print("** Error calculating shortest path (primary)")
        raise
    
    
    primary = {}
    primary["pia_intersect"] = [ px, py ]
    primary["wm_intersect"] = [ wx, wy ]
    primary["soma_center"] = [ soma_x, soma_y ]
    primary["soma_depth_um"] = depth
    try:
        primary["soma_depth_relative"] = depth / raw_thickness
    except:
        primary["soma_depth_relative"] = -1.0   # NaN is not friendly to ruby
    primary["cort_thickness_um"] = raw_thickness
    primary["theta"] = theta
    
    try:
        scale = raw_thickness / blk_thickness
    except:
        scale = -1.0    # NaN is not ruby-friendly
    
    soma_coords_avail = False
    if "swc_file" in jin:
        # if SWC file available, extract soma position from it
        try:
            nrn = swc.read_swc(jin["swc_file"])
            root = nrn.soma_root()
            soma_x = root.x
            soma_y = root.y
            soma_z = root.z
            soma_coords_avail = True
        except:
            # treat this as a fatal error -- if SWC was specified then
            #   it should be used
            print("**** Error reading SWC file '%s'" % jin["swc_file"])
            raise

    if not soma_coords_avail:
        # hope that the 63x data is available. As of May 2017, this seems
        #   to no longer be supplied to the module, but just in case...
        # IT-14567 continue if 63x data not available
        try:
            info = jin["soma_63x"]
            sx, sy = convert_coords_str(info["path_63x"])
            soma_x, soma_y = calculate_centroid(sx, sy)
            soma_x *= float(info["resolution"])
            soma_y *= float(info["resolution"])
            soma_z = float(info["idx"]) * float(info["thickness"])
            soma_coords_avail = True
        except:
            print("** Error reading soma 63x info from input json")
            print("** Translation component of affine matrix is invalid **")

    if not soma_coords_avail:
        raise Exception("** Error: Unable to construct translation component of affine")
    
    try:
        # apply affine rotation to soma position
        translate_x = soma_x*tr_rot[0] + soma_y*tr_rot[1] + soma_z*tr_rot[2]
        translate_y = soma_x*tr_rot[3] + soma_y*tr_rot[4] + soma_z*tr_rot[5]
        translate_z = soma_x*tr_rot[6] + soma_y*tr_rot[7] + soma_z*tr_rot[8]
        # apply translation vector to transform
        tr_rot[ 9] = -translate_x
        tr_rot[10] = -translate_y - depth
        tr_rot[11] = -translate_z
    except:
        print("** Error calculating affine tranform (math fault?)")
        raise
    soma_x = -translate_x
    soma_y = -translate_y - depth
    soma_z = -translate_z
    
    # apply affine rotation to soma position
    translate_x = soma_x*inv_tr_rot[0] + soma_y*inv_tr_rot[1] + soma_z*inv_tr_rot[2]
    translate_y = soma_x*inv_tr_rot[3] + soma_y*inv_tr_rot[4] + soma_z*inv_tr_rot[5]
    translate_z = soma_x*inv_tr_rot[6] + soma_y*inv_tr_rot[7] + soma_z*inv_tr_rot[8]
    
    inv_tr_rot[ 9] = -translate_x
    inv_tr_rot[10] = -translate_y
    inv_tr_rot[11] = -translate_z
    
    try:
        # upright transform. based on rotation in 20x image
        upright = {}
        for i in range(12):
            upright["tvr_%02d" % i] = tr_rot[i]
            upright["trv_%02d" % i] = inv_tr_rot[i]
        jout = {}
        jout["primary"] = primary
        #jout["blockface"] = blockface  # per IT-14567, disable blockface
        jout["upright"] = upright
        alignment = {}
        alignment["scale"] = scale
        alignment["rotate_x"] = theta
        alignment["rotate_y"] = theta
        alignment["rotate_z"] = 0.0
        alignment["scale_x"] = 1.0
        alignment["scale_y"] = 1.0
        alignment["scale_z"] = 1.0
        alignment["skew_x"] = 0.0
        alignment["skew_y"] = 0.0
        alignment["skew_z"] = 0.0
        jout["alignment"] = alignment
    except:
        print("** Internal error **")
        raise
    
    return jout

    #
    # test transform -- bar.swc should match the source file
    #print("source swc: " + jin["swc_file"])
    #print tr_rot
    #morph2 = swc.read_swc(jin["swc_file"])
    #morph2.apply_affine(tr_rot)
    #morph2.save("foo.swc")
    #morph3 = swc.read_swc("foo.swc")
    #morph3.apply_affine(inv_tr_rot)
    #morph3.save("bar.swc")

if __name__ == "__main__":
    # read module input. PipelineModule object automatically parses the 
    #   command line to pull out input.json and output.json file names
    module = PipelineModule()
    jin = module.input_data()   # loads input.json
    jout = main(jin)
    module.write_output_data(jout)  # writes output.json




import numpy as np

def measure_vm(seg):
    vals = np.copy(seg)
    if len(vals) < 1:
        return 0, 0
    mean = np.mean(vals)
    vals -= mean
    rms = np.sqrt(np.mean(np.square(vals)))
    return mean, rms

########################################################################
# experiment-level metrics

def measure_blowout(v, idx0):
    return 1e3 * np.mean(v[idx0:])

def measure_electrode_0(curr, hz, t=0.005):
    n_time_steps = int(t * hz)
    # electrode 0 is the average current reading with zero voltage input
    # (ie, the equivalent of resting potential in current-clamp mode)
    return 1e12 * np.mean(curr[0:n_time_steps])

def measure_seal(v, curr, hz):
    t = np.arange(len(v)) / hz
    return 1e-9 * get_r_from_stable_pulse_response(v, curr, t)

def measure_input_resistance(v, curr, hz):
    t = np.arange(len(v)) / hz
    return 1e-6 * get_r_from_stable_pulse_response(v, curr, t)

def measure_initial_access_resistance(v, curr, hz):
    t = np.arange(len(v)) / hz
    return 1e-6 * get_r_from_peak_pulse_response(v, curr, t)


########################################################################

def get_r_from_stable_pulse_response(v, i, t):
    dv = np.diff(v)
    up_idx = np.flatnonzero(dv > 0)
    down_idx = np.flatnonzero(dv < 0)
#    print up_idx
#    print down_idx
#    print "-----"
    dt = t[1] - t[0]
    one_ms = int(0.001 / dt)
    r = []
    for ii in range(len(up_idx)):
        # take average v and i one ms before
        end = up_idx[ii] - 1
        start = end - one_ms
#        print "\tbase"
#        print "base interval: %d -> %d" % (start, end)
        avg_v_base = np.mean(v[start:end])
        avg_i_base = np.mean(i[start:end])
#        print "\tv: %g" % avg_v_base
#        print "\ti: %g" % avg_i_base
        # take average v and i one ms before end
        end = down_idx[ii]-1
        start = end - one_ms
#        print "\tsteady"
#        print "steady interval: %d -> %d" % (start, end)
        avg_v_steady = np.mean(v[start:end])
        avg_i_steady = np.mean(i[start:end])
#        print "\tv: %g" % avg_v_steady
#        print "\ti: %g" % avg_i_steady
        r_instance = (avg_v_steady-avg_v_base) / (avg_i_steady-avg_i_base)
#        print 1e-6*r_instance
        r.append(r_instance)
    return np.mean(r)

def get_r_from_peak_pulse_response(v, i, t):
    dv = np.diff(v)
    up_idx = np.flatnonzero(dv > 0)
    down_idx = np.flatnonzero(dv < 0)
    dt = t[1] - t[0]
    one_ms = int(0.001 / dt)
    r = []
    for ii in range(len(up_idx)):
        # take average v and i one ms before
        end = up_idx[ii] - 1
        start = end - one_ms
        avg_v_base = np.mean(v[start:end])
        avg_i_base = np.mean(i[start:end])
        # take average v and i one ms before end
        start = up_idx[ii]
        end = down_idx[ii] - 1
        idx = start + np.argmax(i[start:end])
        avg_v_peak = v[idx]
        avg_i_peak = i[idx]
        r_instance = (avg_v_peak-avg_v_base) / (avg_i_peak-avg_i_base)
        r.append(r_instance)
    return np.mean(r)





def get_last_vm_epoch(idx1, stim, hz):
    return idx1-int(0.500 * hz), idx1

def get_first_vm_noise_epoch(idx0, stim, hz):
    t0 = idx0
    t1 = t0 + int(0.0015 * hz)
    return t0, t1

def get_last_vm_noise_epoch(idx1, stim, hz):
    return idx1-int(0.0015 * hz), idx1

#def get_stability_vm_epoch(idx0, stim, hz):
def get_stability_vm_epoch(idx0, stim_start, hz):
    dur = int(0.500 * hz)
    #stim_start = find_stim_start(idx0, stim)
    if dur > stim_start-1:
        dur = stim_start-1
    elif dur <= 0:
        return 0, 0
    return stim_start-1-dur, stim_start-1

def find_stim_start(idx0, stim):
    # find stim start, using adaptation of nathan's numpy algorithm
    di = np.diff(stim)
    up_idx = np.flatnonzero(di > 0)
    down_idx = np.flatnonzero(di < 0)
    first = -1
    for i in range(len(up_idx)):
        if up_idx[i] >= idx0:
            first = up_idx[i]
            break
    for i in range(len(down_idx)):
        if down_idx[i] >= idx0 and down_idx[i] < first:
            first = down_idx[i]
            break
    # +1 to be first index of stim, not last index of pre-stim
    return first + 1

def find_stim_amplitude_and_duration(idx0, stim, hz):

    if len(stim) < idx0:
        idx0 = 0

    stim = stim[idx0:]

    peak_high = max(stim)
    peak_low = min(stim)

    # measure stimulus length
    # find index of first non-zero value, and last return to zero
    nzero = np.where(stim!=0)[0]
    if len(nzero) > 0:
        start = nzero[0]
        end = nzero[-1]
        dur = (end - start) / hz
    else:
        dur = 0
	
    dur = float(dur)

    if abs(peak_high) > abs(peak_low):
        amp = float(peak_high)
    else:
        amp = float(peak_low)

    return amp, dur

def find_stim_interval(idx0, stim, hz):
    stim = stim[idx0:]

    # indices where is the stimulus off
    zero_idxs = np.where(stim == 0)[0]

    # derivative of off indices.  when greater than one, indicates on period 
    dzero_idxs = np.diff(zero_idxs)
    dzero_break_idxs = np.where(dzero_idxs[:] > 1)[0]

    # duration of breaks
    break_durs = dzero_idxs[dzero_break_idxs]

    # indices of breaks
    break_idxs = zero_idxs[dzero_break_idxs] + 1

    # time between break onsets
    dbreaks = np.diff(break_idxs)

    if len(np.unique(break_durs)) == 1 and len(np.unique(dbreaks)) == 1:
        return dbreaks[0] / hz

    return None
